#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform functional tests to verify backward compatibility."""
import os
import shutil
import subprocess

from pymongo import MongoClient
import pytest

from orion.client import create_experiment
from orion.core.io.database import Database
from orion.core.io.database.mongodb import MongoDB
from orion.core.io.database.pickleddb import PickledDB
import orion.core.io.experiment_builder as experiment_builder
from orion.storage.base import get_storage, Storage
from orion.storage.legacy import Legacy


DIRNAME = os.path.dirname(os.path.abspath(__file__))

PYTHON_SCRIPT_PATH = os.path.join(DIRNAME, 'python_api.py')
SCRIPT_PATH = os.path.join(DIRNAME, 'black_box.py')
CONFIG_FILE = os.path.join(DIRNAME, 'random.yaml')

# Ignore pre-0.1.6 because was on orion.core and pypi project was deleted.
VERSIONS = ['0.1.6', '0.1.7', '0.1.8']


def get_branch_argument(version):
    """Get argument to branch.

    Before v0.1.8 it was --branch. From v0.1.8 and forward it is now --branch-to.
    """
    return '--branch' if version < '0.1.8' else '--branch-to'


def clean_mongodb():
    """Clean collections."""
    client = MongoClient(username='user', password='pass', authSource='orion_test')
    database = client.orion_test
    database.experiments.drop()
    database.lying_trials.drop()
    database.trials.drop()
    database.workers.drop()
    database.resources.drop()
    client.close()


def set_env():
    """Set the env vars for the db.

    Notes
    -----
    This function expects the env var ORION_DB_TYPE to be set.

    """
    db_type = os.environ['ORION_DB_TYPE']
    if db_type == 'pickleddb':
        file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'save.pkl')
        os.environ['ORION_DB_ADDRESS'] = file_path
        if os.path.exists(file_path):
            os.remove(file_path)
    elif db_type == 'mongodb':
        os.environ['ORION_DB_ADDRESS'] = 'mongodb://user:pass@localhost'
        os.environ['ORION_DB_NAME'] = 'orion_test'
        clean_mongodb()


def execute(command):
    """Execute a command and return the decoded stdout."""
    out = subprocess.check_output(command.split())
    return out.decode('utf-8')


def setup_virtualenv(version):
    """Create a virtualenv and install Oríon in it for the given version"""
    virtualenv_dir = get_virtualenv_dir(version)
    if os.path.exists(virtualenv_dir):
        shutil.rmtree(virtualenv_dir)
    execute('virtualenv {}'.format(virtualenv_dir))
    command = '{}/bin/pip install --ignore-installed orion=={}'.format(
        virtualenv_dir, version)
    execute(command)


def get_version(orion_script):
    """Get Oríon version for given Oríon executer's path."""
    command = '{} --version'.format(orion_script)
    stdout = subprocess.check_output(command, shell=True)

    return stdout.decode('utf-8').strip('\n')


def get_virtualenv_dir(version):
    """Get virtualenv directory for given version."""
    return 'version-{}'.format(version)


def fill_from_cmdline_api(orion_script, version):
    """Add experiments and trials using the commandline API"""
    # TODO(v0.3.0): Adapt when removing init_only after deprecation phase
    print(execute(' '.join([
        orion_script, '-vv', 'init_only', '--name', 'init-cmdline',
        '--config', CONFIG_FILE,
        SCRIPT_PATH, '-x~uniform(-50,50)'])))

    # TODO(v0.3.0): Adapt when removing init_only after deprecation phase
    print(execute(' '.join([
        orion_script, '-vv', 'init_only', '--name', 'init-cmdline',
        get_branch_argument(version), 'init-cmdline-branch-old',
        '--config', CONFIG_FILE])))

    print(execute(' '.join([
        orion_script, '-vv', 'hunt', '--name', 'hunt-cmdline',
        '--config', CONFIG_FILE,
        SCRIPT_PATH, '-x~uniform(-50,50)'])))

    print(execute(' '.join([
        orion_script, '-vv', 'hunt', '--name', 'hunt-cmdline',
        get_branch_argument(version), 'hunt-cmdline-branch-old',
        '--config', CONFIG_FILE])))


def fill_from_python_api(python_script, version):
    """Add experiments and trials using the python API"""
    print(execute(' '.join([
        python_script, PYTHON_SCRIPT_PATH, version])))


@pytest.fixture(scope='class', params=VERSIONS)
def fill_db(request):
    """Add experiments and trials in DB for given version of Oríon."""
    set_env()

    version = request.param

    setup_virtualenv(version)

    orion_script = os.path.join(get_virtualenv_dir(version), 'bin', 'orion')
    python_script = os.path.join(get_virtualenv_dir(version), 'bin', 'python')

    orion_version = get_version(orion_script)
    assert orion_version == 'orion {}'.format(version)

    fill_from_cmdline_api(orion_script, version)
    if version > '0.1.7':
        fill_from_python_api(python_script, version)

    orion_version = get_version('orion')
    assert orion_version != 'orion {}'.format(version)

    print(execute('orion -vv db upgrade -f'))

    return version


def null_db_instances():
    """Nullify singleton instance so that we can assure independent instantiation tests."""
    Storage.instance = None
    Legacy.instance = None
    Database.instance = None
    MongoDB.instance = None
    PickledDB.instance = None


def build_storage():
    """Build storage from scratch"""
    null_db_instances()
    experiment_builder.setup_storage()

    return get_storage()


@pytest.mark.usefixtures('fill_db')
class TestBackwardCompatibility:
    """Tests for backward compatibility between Oríon versions."""

    def test_db_upgrade(self):
        """Verify db upgrade was successful"""
        storage = build_storage()

        index_info = storage._db.index_information('experiments')
        assert (set([key for key, is_unique in index_info.items() if is_unique]) ==
                set(['_id_', 'name_1_version_1']))

        experiments = storage.fetch_experiments({})
        assert 'version' in experiments[0]
        assert 'priors' in experiments[0]['metadata']

    def test_db_test(self):
        """Verify db test command"""
        out = execute('orion db test')
        assert 'Failure' not in out

    def test_list(self, fill_db):
        """Verify list command"""
        out = execute('orion list')
        assert 'init-cmdline-v1' in out
        assert 'init-cmdline-branch-old-v1' in out
        assert 'hunt-cmdline-v1' in out
        assert 'hunt-cmdline-branch-old-v1' in out

        version = fill_db
        if version > '0.1.7':
            assert 'hunt-python-v1' in out

    def test_status(self, fill_db):
        """Verify status command"""
        out = execute('orion status')
        assert 'init-cmdline-v1' in out
        assert 'init-cmdline-branch-old-v1' in out
        assert 'hunt-cmdline-v1' in out
        assert 'hunt-cmdline-branch-old-v1' in out

        version = fill_db
        if version > '0.1.7':
            assert 'hunt-python-v1' in out

    def test_info_cmdline_api(self):
        """Verify info command from commandline api"""
        out = execute('orion info --name hunt-cmdline')
        assert 'name: hunt-cmdline' in out

    def test_info_python_api(self, fill_db):
        """Verify info command from python api"""
        version = fill_db
        if version < '0.1.8':
            pytest.skip("Python API not supported by {}".format(version))

        out = execute('orion info --name hunt-python')
        assert 'name: hunt-python' in out

    # TODO(v0.3.0): Adapt when removing init_only after deprecation phase
    def test_init_only(self):
        """Verify init_only command"""
        print(execute(' '.join([
            'orion', 'init_only', '--name', 'init-cmdline',
            '--branch-to', 'init-cmdline-branch'])))

    def test_hunt_cmdline_api(self):
        """Verify hunt command from cmdline api parent"""
        print(execute(' '.join([
            'orion', 'hunt', '--name', 'hunt-cmdline',
            '--branch-to', 'hunt-cmdline-branch'])))

    def test_hunt_python_api(self, fill_db):
        """Verify hunt command from python api parent"""
        version = fill_db
        if version < '0.1.8':
            pytest.skip("Python API not supported by {}".format(version))

        def function(x):
            """Evaluate partial information of a quadratic."""
            z = x - 34.56789
            return [dict(
                name='example_objective',
                type='objective',
                value=4 * z**2 + 23.4)]

        exp = create_experiment('hunt-python', branching={'branch-to': 'hunt-python-branch'})
        exp.workon(function, max_trials=10)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Simple one dimensional example for a possible user's script."""

import argparse

from orion.client import report_results


def function(x, z):
    """Evaluate partial information of a quadratic."""
    y = x + z - 34.56789
    return 4 * y**2 + 23.4, 8 * y


def execute():
    """Execute a simple pipeline as an example."""
    # 1. Receive inputs as you want
    parser = argparse.ArgumentParser()
    parser.add_argument('-x', type=float, required=True)
    parser.add_argument('-z', type=float, default=0.0)
    inputs = parser.parse_args()

    # 2. Perform computations
    y, dy = function(inputs.x, inputs.z)

    # 3. Gather and report results
    results = list()
    results.append(dict(
        name='example_objective',
        type='objective',
        value=y))
    results.append(dict(
        name='example_gradient',
        type='gradient',
        value=[dy]))

    report_results(results)


if __name__ == "__main__":
    execute()



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.io.database.mongodb` -- Wrapper for MongoDB
============================================================

.. module:: database
   :platform: Unix
   :synopsis: Implement :class:`orion.core.io.database.AbstractDB` for MongoDB.

"""
import functools

import pymongo

from orion.core.io.database import (
    AbstractDB, DatabaseError, DatabaseTimeout, DuplicateKeyError)


AUTH_FAILED_MESSAGES = [
    "auth failed",
    "Authentication failed."]

INDEX_OP_ERROR_MESSAGES = [
    "index not found with name"]

DUPLICATE_KEY_MESSAGES = [
    "duplicate key error"]


def mongodb_exception_wrapper(method):
    """Convert pymongo exceptions to generic exception types defined in src.core.io.database.

    Current exception types converted:
    pymongo.errors.DuplicateKeyError -> DuplicateKeyError
    pymongo.errors.BulkWriteError[DUPLICATE_KEY_MESSAGES] -> DuplicateKeyError
    pymongo.errors.ConnectionFailure -> DatabaseError
    pymongo.errors.OperationFailure(AUTH_FAILED_MESSAGES) -> DatabaseError

    """
    @functools.wraps(method)
    def _decorator(self, *args, **kwargs):

        try:
            rval = method(self, *args, **kwargs)
        except pymongo.errors.ExecutionTimeout as e:
            # Raised when a database operation times out, exceeding the $maxTimeMS set in
            # the query or command option.
            raise DatabaseTimeout() from e
        except pymongo.errors.NetworkTimeout as e:
            # An operation on an open connection exceeded socketTimeoutMS.
            #
            # The remaining connections in the pool stay open. In the case of a
            # write operation, you cannot know whether it succeeded or failed.
            raise DatabaseTimeout() from e
        except pymongo.errors.WTimeoutError as e:
            # Raised when a database operation times out (i.e. wtimeout expires)
            # before replication completes.
            raise DatabaseTimeout() from e
        except pymongo.errors.DuplicateKeyError as e:
            raise DuplicateKeyError(str(e)) from e
        except pymongo.errors.BulkWriteError as e:
            for error in e.details['writeErrors']:
                if any(m in error["errmsg"] for m in DUPLICATE_KEY_MESSAGES):
                    raise DuplicateKeyError(error["errmsg"]) from e

            raise
        except pymongo.errors.ConnectionFailure as e:
            raise DatabaseError("Connection Failure: database not found on "
                                "specified uri") from e
        except pymongo.errors.OperationFailure as e:
            if any(m in str(e) for m in AUTH_FAILED_MESSAGES):
                raise DatabaseError("Authentication Failure: bad credentials") from e
            elif any(m in str(e) for m in INDEX_OP_ERROR_MESSAGES):
                raise DatabaseError(str(e)) from e
            raise

        return rval

    return _decorator


# pylint: disable=too-many-public-methods
class MongoDB(AbstractDB):
    """Wrap MongoDB with three primary methods `read`, `write`, `remove`.

    Attributes
    ----------
    host : str
       Hostname or MongoDB compliant full credentials+address+database
       specification.

    Information on MongoDB `connection string
    <https://docs.mongodb.com/manual/reference/connection-string/>`_.

    .. seealso:: :class:`orion.core.io.database.AbstractDB` for more on attributes.

    """

    def __init__(self, host='localhost', name=None,
                 port=None, username=None, password=None, serverSelectionTimeoutMS=5000):
        """Init method, see attributes of :class:`AbstractDB`."""
        self.uri = None

        if port is not None:
            port = int(port)
        else:
            port = pymongo.MongoClient.PORT

        super(MongoDB, self).__init__(host, name, port, username, password,
                                      serverSelectionTimeoutMS=serverSelectionTimeoutMS,
                                      authSource=name)

    @mongodb_exception_wrapper
    def initiate_connection(self):
        """Connect to database, unless MongoDB `is_connected`.

        :raises :exc:`DatabaseError`: if connection or authentication fails

        """
        if self.is_connected:
            return

        self._sanitize_attrs()

        self._conn = pymongo.MongoClient(self.uri if self.uri else self.host,
                                         port=self.port,
                                         username=self.username,
                                         password=self.password,
                                         **self.options)
        self._db = self._conn[self.name]
        self._db.command('ismaster')  # .. seealso:: :meth:`is_connected`

    @property
    def is_connected(self):
        """True, if practical connection has been achieved.

        .. note:: MongoDB does not do this automatically when creating the client.
        """
        try:
            self._db.command('ismaster')
        except (pymongo.errors.ConnectionFailure,
                pymongo.errors.OperationFailure,
                TypeError, AttributeError):
            _is_connected = False
        else:
            _is_connected = True
        return _is_connected

    def close_connection(self):
        """Disconnect from database.

        .. note:: Doesn't really do anything because MongoDB reopens connection,
           when a client or client-derived object is accessed.
        """
        self._conn.close()

    def ensure_index(self, collection_name, keys, unique=False):
        """Create given indexes if they do not already exist in database.

        .. seealso:: :meth:`AbstractDB.ensure_index` for argument documentation.

        """
        # MongoDB's `create_index()` is idempotent, which means it will only
        # create new indexes if they do not already exists. That's why we do
        # not need to verify if indexes already exists.
        dbcollection = self._db[collection_name]

        keys = self._convert_index_keys(keys)

        dbcollection.create_index(keys, unique=unique, background=True)

    def index_information(self, collection_name):
        """Return dict of names and sorting order of indexes"""
        dbcollection = self._db[collection_name]
        return {index: specs.get('unique', False) or index == '_id_'
                for index, specs in dbcollection.index_information().items()}

    @mongodb_exception_wrapper
    def drop_index(self, collection_name, name):
        """Remove index from the database"""
        dbcollection = self._db[collection_name]
        dbcollection.drop_index(name)

    def _convert_index_keys(self, keys):
        """Convert index keys to MongoDB ones."""
        if not isinstance(keys, (list, tuple)):
            keys = [(keys, self.ASCENDING)]

        converted_keys = []
        for key, sort_order in keys:
            converted_keys.append((key, self._convert_sort_order(sort_order)))

        return converted_keys

    def _convert_sort_order(self, sort_order):
        """Convert generic `AbstractDB` sort orders to MongoDB ones."""
        if sort_order is self.ASCENDING:
            return pymongo.ASCENDING
        elif sort_order is self.DESCENDING:
            return pymongo.DESCENDING
        else:
            raise RuntimeError("Invalid database sort order %s" %
                               str(sort_order))

    @mongodb_exception_wrapper
    def write(self, collection_name, data, query=None):
        """Write new information to a collection. Perform insert or update.

        .. seealso:: :meth:`AbstractDB.write` for argument documentation.

        """
        dbcollection = self._db[collection_name]

        if query is None:
            # We can assume that we do not want to update.
            # So we do insert_many instead.
            if type(data) not in (list, tuple):
                data = [data]

            result = dbcollection.insert_many(documents=data)
            return len(result.inserted_ids)

        update_data = {'$set': data}

        result = dbcollection.update_many(filter=query,
                                          update=update_data,
                                          upsert=False)
        return result.modified_count

    def read(self, collection_name, query=None, selection=None):
        """Read a collection and return a value according to the query.

        .. seealso:: :meth:`AbstractDB.read` for argument documentation.

        """
        dbcollection = self._db[collection_name]

        cursor = dbcollection.find(query, selection)
        dbdocs = list(cursor)

        return dbdocs

    @mongodb_exception_wrapper
    def read_and_write(self, collection_name, query, data, selection=None):
        """Read a collection's document and update the found document.

        Returns the updated document, or None if nothing found.

        .. seealso:: :meth:`AbstractDB.read_and_write` for
                     argument documentation.

        """
        dbcollection = self._db[collection_name]

        update_data = {'$set': data}

        dbdoc = dbcollection.find_one_and_update(
            query, update_data, projection=selection,
            return_document=pymongo.ReturnDocument.AFTER)

        return dbdoc

    def count(self, collection_name, query=None):
        """Count the number of documents in a collection which match the `query`.

        .. seealso:: :meth:`AbstractDB.count` for argument documentation.

        """
        dbcollection = self._db[collection_name]
        if not isinstance(getattr(dbcollection, 'count_documents'), pymongo.collection.Collection):
            return dbcollection.count_documents(filter=query if query else {})

        return dbcollection.count(filter=query)

    def remove(self, collection_name, query):
        """Delete from a collection document[s] which match the `query`.

        .. seealso:: :meth:`AbstractDB.remove` for argument documentation.

        """
        dbcollection = self._db[collection_name]

        result = dbcollection.delete_many(filter=query)
        return result.deleted_count

    def _sanitize_attrs(self):
        """Sanitize attributes using MongoDB's 'uri_parser' module."""
        try:
            # Host can be a valid MongoDB URI
            settings = pymongo.uri_parser.parse_uri(self.host, default_port=self.port)
        except pymongo.errors.InvalidURI:  # host argument was a hostname
            if self.port is None:
                self.port = pymongo.MongoClient.PORT
        else:  # host argument was a URI
            # Arguments in MongoClient overwrite elements from URI

            self.uri = self.host
            self.host, _port = settings['nodelist'][0]
            if settings['database'] is not None:
                self.name = settings['database']
            if _port is not None:
                self.port = _port
            if settings['username'] is not None:
                self.username = settings['username']
            if settings['password'] is not None:
                self.password = settings['password']

            # Use new self.name if authSource not specified in URI
            self.options['authSource'] = settings['options'].get('authsource', self.name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for :mod:`orion.algo.hyperband`."""

import hashlib

import numpy as np
import pytest

from orion.algo.hyperband import Bracket, compute_budgets, Hyperband
from orion.algo.space import Fidelity, Integer, Real, Space


@pytest.fixture
def space():
    """Create a Space with a real dimension and a fidelity value."""
    space = Space()
    space.register(Real('lr', 'uniform', 0, 1))
    space.register(Fidelity('epoch', 1, 9, 3))
    return space


@pytest.fixture
def budgets():
    """Return a configuration for a bracket."""
    return [(9, 1), (3, 3), (1, 9)]


@pytest.fixture
def hyperband(space):
    """Return an instance of Hyperband."""
    return Hyperband(space, repetitions=1)


@pytest.fixture
def bracket(budgets, hyperband):
    """Return a `Bracket` instance configured with `b_config`."""
    return Bracket(hyperband, budgets, 1)


@pytest.fixture
def rung_0():
    """Create fake points and objectives for rung 0."""
    points = np.linspace(0, 8, 9)
    return dict(
        n_trials=9,
        resources=1,
        results={hashlib.md5(str([point]).encode('utf-8')).hexdigest(): (point, (1, point))
                 for point in points})


@pytest.fixture
def rung_1(rung_0):
    """Create fake points and objectives for rung 1."""
    values = map(lambda v: (v[0], (3, v[0])), list(sorted(rung_0['results'].values()))[:3])
    return dict(
        n_trials=3,
        resources=3,
        results={hashlib.md5(str([value[0]]).encode('utf-8')).hexdigest(): value
                 for value in values})


@pytest.fixture
def rung_2(rung_1):
    """Create fake points and objectives for rung 1."""
    values = map(lambda v: (v[0], (9, v[0])), list(sorted(rung_1['results'].values()))[:1])
    return dict(
        n_trials=1,
        resources=9,
        results={hashlib.md5(str([value[0]]).encode('utf-8')).hexdigest(): value
                 for value in values})


def test_compute_budgets():
    """Verify proper computation of budgets on a logarithmic scale"""
    # Check typical values
    assert compute_budgets(81, 3) == [[(81, 1), (27, 3), (9, 9), (3, 27), (1, 81)],
                                      [(27, 3), (9, 9), (3, 27), (1, 81)],
                                      [(9, 9), (3, 27), (1, 81)],
                                      [(6, 27), (2, 81)], [(5, 81)]]
    assert compute_budgets(16, 4) == [[(16, 1), (4, 4), (1, 16)], [(4, 4), (1, 16)], [(3, 16)]]
    assert compute_budgets(16, 5) == [[(5, 3), (1, 16)], [(2, 16)]]


class TestBracket():
    """Tests for the `Bracket` class."""

    def test_rungs_creation(self, bracket):
        """Test the creation of rungs for bracket 0."""
        assert len(bracket.rungs) == 3
        assert bracket.rungs[0] == dict(n_trials=9, resources=1, results=dict())
        assert bracket.rungs[1] == dict(n_trials=3, resources=3, results=dict())
        assert bracket.rungs[2] == dict(n_trials=1, resources=9, results=dict())

    def test_register(self, hyperband, bracket):
        """Check that a point is correctly registered inside a bracket."""
        bracket.hyperband = hyperband
        point = (1, 0.0)
        point_hash = hashlib.md5(str([0.0]).encode('utf-8')).hexdigest()

        bracket.register(point, 0.0)

        assert len(bracket.rungs[0])
        assert point_hash in bracket.rungs[0]['results']
        assert (0.0, point) == bracket.rungs[0]['results'][point_hash]

    def test_bad_register(self, hyperband, bracket):
        """Check that a non-valid point is not registered."""
        bracket.hyperband = hyperband

        with pytest.raises(IndexError) as ex:
            bracket.register((55, 0.0), 0.0)

        assert 'Bad fidelity level 55' in str(ex.value)

    def test_candidate_promotion(self, hyperband, bracket, rung_0):
        """Test that correct point is promoted."""
        bracket.hyperband = hyperband
        bracket.rungs[0] = rung_0

        points = bracket.get_candidates(0)

        assert points[0] == (1, 0.0)

    def test_promotion_with_rung_1_hit(self, hyperband, bracket, rung_0):
        """Test that get_candidate gives us the next best thing if point is already in rung 1."""
        point = (1, 0.0)
        point_hash = hashlib.md5(str([0.0]).encode('utf-8')).hexdigest()
        bracket.hyperband = hyperband
        bracket.rungs[0] = rung_0
        bracket.rungs[1]['results'][point_hash] = (0.0, point)

        points = bracket.get_candidates(0)

        assert points[0] == (1, 1)

    def test_no_promotion_when_rung_full(self, hyperband, bracket, rung_0, rung_1):
        """Test that get_candidate returns `None` if rung 1 is full."""
        bracket.hyperband = hyperband
        bracket.rungs[0] = rung_0
        bracket.rungs[1] = rung_1

        points = bracket.get_candidates(0)

        assert points == []

    def test_no_promotion_if_not_completed(self, hyperband, bracket, rung_0):
        """Test the get_candidate return None if trials are not completed."""
        bracket.hyperband = hyperband
        bracket.rungs[0] = rung_0
        rung = bracket.rungs[0]['results']

        # points = bracket.get_candidates(0)

        for p_id in rung.keys():
            rung[p_id] = (None, rung[p_id][1])

        with pytest.raises(AssertionError):
            bracket.get_candidates(0)

    def test_is_done(self, bracket, rung_0):
        """Test that the `is_done` property works."""
        assert not bracket.is_done

        # Actual value of the point is not important here
        bracket.rungs[2]['results'] = {'1': (1, 0.0), '2': (1, 0.0), '3': (1, 0.0)}

        assert bracket.is_done

    def test_update_rungs_return_candidate(self, hyperband, bracket, rung_1):
        """Check if a valid modified candidate is returned by update_rungs."""
        bracket.hyperband = hyperband
        bracket.rungs[1] = rung_1
        point_hash = hashlib.md5(str([0.0]).encode('utf-8')).hexdigest()

        candidates = bracket.promote()

        assert point_hash in bracket.rungs[1]['results']
        assert bracket.rungs[1]['results'][point_hash] == (0.0, (3, 0.0))
        assert candidates[0][0] == 9

    def test_update_rungs_return_no_candidate(self, hyperband, bracket, rung_1):
        """Check if no candidate is returned by update_rungs."""
        bracket.hyperband = hyperband

        candidate = bracket.promote()

        assert candidate is None

    def test_get_point_max_resource(self, hyperband, bracket, rung_0, rung_1, rung_2):
        """Test to get the max resource R for a particular point"""
        bracket.hyperband = hyperband
        bracket.rungs[0] = rung_0

        assert bracket.get_point_max_resource(point=(1, 0.0)) == 1
        assert bracket.get_point_max_resource(point=(1, 8.0)) == 1

        bracket.rungs[1] = rung_1
        assert bracket.get_point_max_resource(point=(1, 0.0)) == 3
        assert bracket.get_point_max_resource(point=(1, 8.0)) == 1

        bracket.rungs[2] = rung_2
        assert bracket.get_point_max_resource(point=(1, 0.0)) == 9
        assert bracket.get_point_max_resource(point=(1, 8.0)) == 1

    def test_repr(self, bracket, rung_0, rung_1, rung_2):
        """Test the string representation of Bracket"""
        bracket.rungs[0] = rung_0
        bracket.rungs[1] = rung_1
        bracket.rungs[2] = rung_2

        assert str(bracket) == 'Bracket(resource=[1, 3, 9], repetition id=1)'


class TestHyperband():
    """Tests for the algo Hyperband."""

    def test_register(self, hyperband, bracket, rung_0, rung_1):
        """Check that a point is registered inside the bracket."""
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband
        bracket.rungs = [rung_0, rung_1]
        point = (1, 0.0)
        point_hash = hashlib.md5(str([0.0]).encode('utf-8')).hexdigest()

        hyperband.observe([point], [{'objective': 0.0}])

        assert len(bracket.rungs[0])
        assert point_hash in bracket.rungs[0]['results']
        assert (0.0, point) == bracket.rungs[0]['results'][point_hash]

    def test_register_bracket_multi_fidelity(self, space):
        """Check that a point is registered inside the same bracket for diff fidelity."""
        hyperband = Hyperband(space)

        value = 50
        fidelity = 1
        point = (fidelity, value)
        point_hash = hashlib.md5(str([value]).encode('utf-8')).hexdigest()

        hyperband.observe([point], [{'objective': 0.0}])

        bracket = hyperband.brackets[0]

        assert len(bracket.rungs[0])
        assert point_hash in bracket.rungs[0]['results']
        assert (0.0, point) == bracket.rungs[0]['results'][point_hash]

        fidelity = 3
        point = [fidelity, value]
        point_hash = hashlib.md5(str([value]).encode('utf-8')).hexdigest()

        hyperband.observe([point], [{'objective': 0.0}])

        assert len(bracket.rungs[0])
        assert point_hash in bracket.rungs[1]['results']
        assert (0.0, point) != bracket.rungs[0]['results'][point_hash]
        assert (0.0, point) == bracket.rungs[1]['results'][point_hash]

    def test_register_next_bracket(self, space):
        """Check that a point is registered inside the good bracket when higher fidelity."""
        hyperband = Hyperband(space)

        value = 50
        fidelity = 3
        point = (fidelity, value)
        point_hash = hashlib.md5(str([value]).encode('utf-8')).hexdigest()

        hyperband.observe([point], [{'objective': 0.0}])

        assert sum(len(rung['results']) for rung in hyperband.brackets[0].rungs) == 0
        assert sum(len(rung['results']) for rung in hyperband.brackets[1].rungs) == 1
        assert sum(len(rung['results']) for rung in hyperband.brackets[2].rungs) == 0
        assert point_hash in hyperband.brackets[1].rungs[0]['results']
        assert (0.0, point) == hyperband.brackets[1].rungs[0]['results'][point_hash]

        value = 51
        fidelity = 9
        point = (fidelity, value)
        point_hash = hashlib.md5(str([value]).encode('utf-8')).hexdigest()

        hyperband.observe([point], [{'objective': 0.0}])

        assert sum(len(rung['results']) for rung in hyperband.brackets[0].rungs) == 0
        assert sum(len(rung['results']) for rung in hyperband.brackets[1].rungs) == 1
        assert sum(len(rung['results']) for rung in hyperband.brackets[2].rungs) == 1
        assert point_hash in hyperband.brackets[2].rungs[0]['results']
        assert (0.0, point) == hyperband.brackets[2].rungs[0]['results'][point_hash]

    def test_register_invalid_fidelity(self, space):
        """Check that a point cannot registered if fidelity is invalid."""
        hyperband = Hyperband(space)

        value = 50
        fidelity = 2
        point = (fidelity, value)

        with pytest.raises(ValueError) as ex:
            hyperband.observe([point], [{'objective': 0.0}])

        assert 'No bracket found for point' in str(ex.value)

    def test_register_corrupted_db(self, caplog, space):
        """Check that a point cannot registered if passed in order diff than fidelity."""
        hyperband = Hyperband(space)

        value = 50
        fidelity = 3
        point = (fidelity, value)

        hyperband.observe([point], [{'objective': 0.0}])
        assert 'Point registered to wrong bracket' not in caplog.text

        fidelity = 1
        point = [fidelity, value]

        caplog.clear()
        hyperband.observe([point], [{'objective': 0.0}])
        assert 'Point registered to wrong bracket' in caplog.text

    def test_get_id(self, space):
        """Test valid id of points"""
        hyperband = Hyperband(space)

        assert hyperband.get_id(['whatever', 1]) == hyperband.get_id(['is here', 1])
        assert hyperband.get_id(['whatever', 1]) != hyperband.get_id(['is here', 2])
        assert hyperband.get_id(['whatever', 1], ignore_fidelity=False) != \
            hyperband.get_id(['is here', 1], ignore_fidelity=False)
        assert hyperband.get_id(['whatever', 1], ignore_fidelity=False) != \
            hyperband.get_id(['is here', 2], ignore_fidelity=False)
        assert hyperband.get_id(['same', 1], ignore_fidelity=False) == \
            hyperband.get_id(['same', 1], ignore_fidelity=False)
        assert hyperband.get_id(['same', 1], ignore_fidelity=False) != \
            hyperband.get_id(['same', 1])

    def test_get_id_multidim(self):
        """Test valid id for points with dim of shape > 1"""
        space = Space()
        space.register(Fidelity('epoch', 1, 9, 3))
        space.register(Real('lr', 'uniform', 0, 1, shape=2))

        hyperband = Hyperband(space)

        assert hyperband.get_id(['whatever', [1, 1]]) == hyperband.get_id(['is here', [1, 1]])
        assert hyperband.get_id(['whatever', [1, 1]]) != hyperband.get_id(['is here', [2, 2]])
        assert hyperband.get_id(['whatever', [1, 1]], ignore_fidelity=False) != \
            hyperband.get_id(['is here', [1, 1]], ignore_fidelity=False)
        assert hyperband.get_id(['whatever', [1, 1]], ignore_fidelity=False) != \
            hyperband.get_id(['is here', [2, 2]], ignore_fidelity=False)
        assert hyperband.get_id(['same', [1, 1]], ignore_fidelity=False) == \
            hyperband.get_id(['same', [1, 1]], ignore_fidelity=False)
        assert hyperband.get_id(['same', [1, 1]], ignore_fidelity=False) != \
            hyperband.get_id(['same', [1, 1]])

    def test_suggest_new(self, monkeypatch, hyperband, bracket, rung_0, rung_1, rung_2):
        """Test that a new point is sampled."""
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband

        def sample(num=1, seed=None):
            return [('fidelity', i) for i in range(num)]

        monkeypatch.setattr(hyperband.space, 'sample', sample)

        points = hyperband.suggest()

        assert points[0] == (1.0, 0)
        assert points[1] == (1.0, 1)

    def test_suggest_duplicates_between_calls(self, monkeypatch, hyperband, bracket):
        """Test that same points are not allowed in different suggest call of
        the same hyperband execution.
        """
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband

        duplicate_point = ('fidelity', 0.0)
        new_point = ('fidelity', 0.5)

        duplicate_id = hashlib.md5(str([duplicate_point]).encode('utf-8')).hexdigest()
        bracket.rungs[0]['results'] = {duplicate_id: (0.0, duplicate_point)}

        hyperband.trial_info_wo_fidelity[hyperband.get_id(duplicate_point)] = bracket

        points = [duplicate_point, new_point]

        def sample(num=1, seed=None):
            return points + [('fidelity', i) for i in range(num - 2)]

        monkeypatch.setattr(hyperband.space, 'sample', sample)

        assert hyperband.suggest()[0][1] == new_point[1]

    def test_suggest_duplicates_one_call(self, monkeypatch, hyperband, bracket):
        """Test that same points are not allowed in the same suggest call ofxs
        the same hyperband execution.
        """
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband

        zhe_point = [(1, 0.0), (1, 1.0), (1, 1.0), (1, 2.0)]

        def sample(num=1, seed=None):
            return zhe_point * num

        monkeypatch.setattr(hyperband.space, 'sample', sample)
        zhe_samples = hyperband.suggest()

        assert zhe_samples[0][1] == 0.0
        assert zhe_samples[1][1] == 1.0
        assert zhe_samples[2][1] == 2.0

        zhe_point = [(3, 0.0), (3, 1.0), (3, 5.0), (3, 4.0)]
        hyperband.trial_info_wo_fidelity[hyperband.get_id((1, 0.0))] = bracket
        hyperband.trial_info_wo_fidelity[hyperband.get_id((1, 1.0))] = bracket
        zhe_samples = hyperband.suggest()
        assert zhe_samples[0][1] == 5.0
        assert zhe_samples[1][1] == 4.0

    def test_suggest_duplicates_between_execution(self, monkeypatch, hyperband, budgets):
        """Test that sampling collisions are handled between different hyperband execution."""
        hyperband.repetitions = 2
        bracket = Bracket(hyperband, budgets, 1)
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband

        for i in range(9):
            hyperband.observe([(1, i)], [{'objective': i}])

        for i in range(3):
            hyperband.observe([(3, i)], [{'objective': i}])

        hyperband.observe([(9, 0)], [{'objective': 0}])

        assert not hyperband.is_done

        zhe_point = [(9, 0), (9, 1), (9, 2)]

        def sample(num=1, seed=None):
            return zhe_point * num

        monkeypatch.setattr(hyperband.space, 'sample', sample)
        zhe_samples = hyperband.suggest()
        assert zhe_samples == [(9, 1), (9, 2)]

    def test_suggest_inf_duplicates(self, monkeypatch, hyperband, bracket, rung_0, rung_1, rung_2):
        """Test that sampling inf collisions will return None."""
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband

        zhe_point = ('fidelity', 0.0)
        hyperband.trial_info_wo_fidelity[hyperband.get_id(zhe_point)] = bracket

        def sample(num=1, seed=None):
            return [zhe_point] * num

        monkeypatch.setattr(hyperband.space, 'sample', sample)

        assert hyperband.suggest() is None

    def test_suggest_in_finite_cardinality(self):
        """Test that suggest None when search space is empty"""
        space = Space()
        space.register(Integer('yolo1', 'uniform', 0, 6))
        space.register(Fidelity('epoch', 1, 9, 3))

        hyperband = Hyperband(space, repetitions=1)
        for i in range(6):
            hyperband.observe([(1, i)], [{'objective': i}])

        assert hyperband.suggest() is None

    def test_suggest_promote(self, hyperband, bracket, rung_0):
        """Test that correct point is promoted and returned."""
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband
        bracket.rungs[0] = rung_0

        points = hyperband.suggest()

        assert points == [(3, i) for i in range(3)]

    def test_is_filled(self, hyperband, bracket, rung_0, rung_1, rung_2):
        """Test that Hyperband bracket detects when rung is filled."""
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband
        bracket.rungs[0] = rung_0

        rung = bracket.rungs[0]['results']
        trial_id = next(iter(rung.keys()))
        objective, point = rung.pop(trial_id)

        assert not bracket.is_filled
        assert not bracket.has_rung_filled(0)

        rung[trial_id] = (objective, point)

        assert bracket.is_filled
        assert bracket.has_rung_filled(0)
        assert not bracket.has_rung_filled(1)
        assert not bracket.has_rung_filled(2)

        bracket.rungs[1] = rung_1

        rung = bracket.rungs[1]['results']
        trial_id = next(iter(rung.keys()))
        objective, point = rung.pop(trial_id)

        assert bracket.is_filled  # Should depend first rung only
        assert bracket.has_rung_filled(0)
        assert not bracket.has_rung_filled(1)

        rung[trial_id] = (objective, point)

        assert bracket.is_filled  # Should depend first rung only
        assert bracket.has_rung_filled(0)
        assert bracket.has_rung_filled(1)
        assert not bracket.has_rung_filled(2)

        bracket.rungs[2] = rung_2

        rung = bracket.rungs[2]['results']
        trial_id = next(iter(rung.keys()))
        objective, point = rung.pop(trial_id)

        assert bracket.is_filled  # Should depend first rung only
        assert bracket.has_rung_filled(0)
        assert bracket.has_rung_filled(1)
        assert not bracket.has_rung_filled(2)

        rung[trial_id] = (objective, point)

        assert bracket.is_filled  # Should depend first rung only
        assert bracket.has_rung_filled(0)
        assert bracket.has_rung_filled(1)
        assert bracket.has_rung_filled(2)

    def test_is_ready(self, hyperband, bracket, rung_0, rung_1, rung_2):
        """Test that Hyperband bracket detects when rung is ready."""
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband
        bracket.rungs[0] = rung_0

        rung = bracket.rungs[0]['results']
        trial_id = next(iter(rung.keys()))
        objective, point = rung[trial_id]
        rung[trial_id] = (None, point)

        assert not bracket.is_ready()
        assert not bracket.is_ready(0)

        rung[trial_id] = (objective, point)

        assert bracket.is_ready()
        assert bracket.is_ready(0)
        assert not bracket.is_ready(1)
        assert not bracket.is_ready(2)

        bracket.rungs[1] = rung_1

        rung = bracket.rungs[1]['results']
        trial_id = next(iter(rung.keys()))
        objective, point = rung[trial_id]
        rung[trial_id] = (None, point)

        assert not bracket.is_ready()  # Should depend on last rung that contains trials
        assert bracket.is_ready(0)
        assert not bracket.is_ready(1)
        assert not bracket.is_ready(2)

        rung[trial_id] = (objective, point)

        assert bracket.is_ready()  # Should depend on last rung that contains trials
        assert bracket.is_ready(0)
        assert bracket.is_ready(1)
        assert not bracket.is_ready(2)

        bracket.rungs[2] = rung_2

        rung = bracket.rungs[2]['results']
        trial_id = next(iter(rung.keys()))
        objective, point = rung[trial_id]
        rung[trial_id] = (None, point)

        assert not bracket.is_ready()  # Should depend on last rung that contains trials
        assert bracket.is_ready(0)
        assert bracket.is_ready(1)
        assert not bracket.is_ready(2)

        rung[trial_id] = (objective, point)

        assert bracket.is_ready()  # Should depend on last rung that contains trials
        assert bracket.is_ready(0)
        assert bracket.is_ready(1)
        assert bracket.is_ready(2)

    def test_suggest_opt_out(self, hyperband, bracket, rung_0, rung_1, rung_2):
        """Test that Hyperband opts out when rungs are not ready."""
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband

        bracket.rungs[0] = rung_0

        trial_id = next(iter(rung_0['results'].keys()))
        objective, point = rung_0['results'][trial_id]
        rung_0['results'][trial_id] = (None, point)

        points = hyperband.suggest()

        assert points is None

    def test_seed_rng(self, hyperband):
        """Test that algo is seeded properly"""
        hyperband.seed_rng(1)
        a = hyperband.suggest(1)
        # Hyperband will always return the full first rung
        assert np.allclose(a, hyperband.suggest(1))

        hyperband.seed_rng(2)
        assert not np.allclose(a, hyperband.suggest(1))

    def test_set_state(self, hyperband):
        """Test that state is reset properly"""
        hyperband.seed_rng(1)
        state = hyperband.state_dict
        points = hyperband.suggest(1)
        # Hyperband will always return the full first rung
        assert np.allclose(points, hyperband.suggest(1))

        hyperband.seed_rng(2)
        assert not np.allclose(points, hyperband.suggest(1))

        hyperband.set_state(state)
        assert np.allclose(points, hyperband.suggest(1))

    def test_full_process(self, monkeypatch, hyperband):
        """Test Hyperband full process."""
        points = [('fidelity', i) for i in range(4000)]

        def sample(num=1, seed=None):
            return points[:num]

        monkeypatch.setattr(hyperband.space, 'sample', sample)

        # Fill all brackets' first rung

        for i in range(3):
            point = hyperband.suggest()[0]
            assert point == (9, i)
            hyperband.observe([point], [{'objective': None}])

        for i in range(3):
            point = hyperband.suggest()[0]
            assert point == (3, i + 3)
            hyperband.observe([point], [{'objective': None}])

        for i in range(9):
            point = hyperband.suggest()[0]
            assert point == (1, i + 3 + 3)
            hyperband.observe([point], [{'objective': None}])

        assert hyperband.brackets[0].has_rung_filled(0)
        assert not hyperband.brackets[0].is_ready()
        assert hyperband.suggest() is None
        assert hyperband.suggest() is None

        # Observe first bracket first rung

        for i in range(9):
            hyperband.observe([(1, i + 3 + 3)], [{'objective': 16 - i}])

        assert hyperband.brackets[0].is_ready()
        assert not hyperband.brackets[1].is_ready()
        assert not hyperband.brackets[2].is_ready()

        # Promote first bracket first rung

        for i in range(3):
            point = hyperband.suggest()[0]
            assert point == (3, 3 + 3 + 9 - 1 - i)
            hyperband.observe([point], [{'objective': None}])

        assert hyperband.brackets[0].has_rung_filled(1)
        assert not hyperband.brackets[0].is_ready()
        assert not hyperband.brackets[1].is_ready()
        assert not hyperband.brackets[2].is_ready()

        # Observe first bracket second rung

        for i in range(3):
            hyperband.observe([(3, 3 + 3 + 9 - 1 - i)], [{'objective': 8 - i}])

        assert hyperband.brackets[0].is_ready()
        assert not hyperband.brackets[1].is_ready()
        assert not hyperband.brackets[2].is_ready()

        # Observe second bracket first rung

        for i in range(3):
            hyperband.observe([(3, i + 3)], [{'objective': 8 - i}])

        assert hyperband.brackets[0].is_ready()
        assert hyperband.brackets[1].is_ready()
        assert not hyperband.brackets[2].is_ready()

        # Promote second bracket first rung

        for i in range(1):
            point = hyperband.suggest()[0]
            assert point == (9, 3 + 3 - 1 - i)
            hyperband.observe([point], [{'objective': None}])

        assert hyperband.brackets[0].is_ready()
        assert hyperband.brackets[1].has_rung_filled(1)
        assert not hyperband.brackets[1].is_ready()
        assert not hyperband.brackets[2].is_ready()

        # Observe third bracket first rung

        for i in range(3):
            hyperband.observe([(9, i)], [{'objective': 3 - i}])

        assert not hyperband.brackets[0].is_ready(2)
        assert not hyperband.brackets[1].is_ready(1)
        assert hyperband.brackets[2].is_ready(0)
        assert hyperband.brackets[2].is_done

        # Observe second bracket second rung

        for i in range(1):
            hyperband.observe([(9, 3 + 3 - 1 - i)], [{'objective': 5 - i}])

        assert not hyperband.brackets[0].is_ready(2)
        assert hyperband.brackets[1].is_ready(1)
        assert hyperband.brackets[1].is_done

        # Observe first bracket third rung

        for i in range(1):
            point = hyperband.suggest()[0]
            assert point == (9, 3 + 3 + 9 - 1 - 2 + i)
            hyperband.observe([point], [{'objective': 3 - i}])

        assert hyperband.is_done
        assert hyperband.brackets[0].is_ready(2)
        assert hyperband.brackets[0].is_done
        assert hyperband.suggest() is None

        # Refresh repeat and execution times property
        monkeypatch.setattr(hyperband, 'repetitions', 2)
        monkeypatch.setattr(hyperband, 'executed_times', 0)
        hyperband.observe([(9, 12)], [{'objective': 3 - i}])
        assert not hyperband.is_done
        assert not hyperband.brackets[0].is_ready(2)
        assert not hyperband.brackets[0].is_done
        assert hyperband.suggest()[0] == (9, 3)



# -*- coding: utf-8 -*-
"""
:mod:`orion.storage.legacy` -- Legacy storage
=============================================

.. module:: legacy
   :platform: Unix
   :synopsis: Old Storage implementation

"""
import datetime
import json
import logging

import orion.core
from orion.core.io.convert import JSONConverter
from orion.core.io.database import Database, OutdatedDatabaseError
import orion.core.utils.backward as backward
from orion.core.utils.exceptions import MissingResultFile
from orion.core.worker.trial import Trial, validate_status
from orion.storage.base import (
    BaseStorageProtocol, FailedUpdate, get_uid, MissingArguments)

log = logging.getLogger(__name__)


def get_database():
    """Return current database

    This is a wrapper around the Database Singleton object to provide
    better error message when it is used without being initialized.

    Raises
    ------
    RuntimeError
        If the underlying database was not initialized prior to calling this function

    Notes
    -----
    To initialize the underlying database you must first call `Database(...)`
    with the appropriate arguments for the chosen backend

    """
    return Database()


def setup_database(config=None):
    """Create the Database instance from a configuration.

    Parameters
    ----------
    config: dict
        Configuration for the database backend. If not defined, global configuration
        is used.

    """
    if config is None:
        # TODO: How could we support orion.core.config.storage.database as well?
        config = orion.core.config.database.to_dict()

    db_opts = config
    dbtype = db_opts.pop('type')

    log.debug("Creating %s database client with args: %s", dbtype, db_opts)
    try:
        Database(of_type=dbtype, **db_opts)
    except ValueError:
        if Database().__class__.__name__.lower() != dbtype.lower():
            raise


class Legacy(BaseStorageProtocol):
    """Legacy protocol, store all experiments and trials inside the Database()

    Parameters
    ----------
    config: Dict
        configuration definition passed from experiment_builder
        to storage factory to legacy constructor.
        See `~orion.io.database.Database` for more details
    setup: bool
        Setup the database (create indexes)

    """

    def __init__(self, database=None, setup=True):
        if database is not None:
            setup_database(database)

        self._db = Database()

        if setup:
            self._setup_db()

    def _setup_db(self):
        """Database index setup"""
        if backward.db_is_outdated(self._db):
            raise OutdatedDatabaseError("The database is outdated. You can upgrade it with the "
                                        "command `orion db upgrade`.")

        self._db.index_information('experiment')
        self._db.ensure_index('experiments',
                              [('name', Database.ASCENDING),
                               ('version', Database.ASCENDING)],
                              unique=True)

        self._db.ensure_index('experiments', 'metadata.datetime')

        self._db.ensure_index('trials', 'experiment')
        self._db.ensure_index('trials', 'status')
        self._db.ensure_index('trials', 'results')
        self._db.ensure_index('trials', 'start_time')
        self._db.ensure_index('trials', [('end_time', Database.DESCENDING)])

    def create_experiment(self, config):
        """See :func:`~orion.storage.BaseStorageProtocol.create_experiment`"""
        return self._db.write('experiments', data=config, query=None)

    def delete_experiment(self, experiment=None, uid=None):
        """See :func:`~orion.storage.BaseStorageProtocol.delete_experiment`"""
        uid = get_uid(experiment, uid)
        return self._db.remove('experiments', query={'_id': uid})

    def update_experiment(self, experiment=None, uid=None, where=None, **kwargs):
        """See :func:`~orion.storage.BaseStorageProtocol.update_experiment`"""
        uid = get_uid(experiment, uid)

        if where is None:
            where = dict()

        if uid is not None:
            where['_id'] = uid
        return self._db.write('experiments', data=kwargs, query=where)

    def fetch_experiments(self, query, selection=None):
        """See :func:`~orion.storage.BaseStorageProtocol.fetch_experiments`"""
        return self._db.read('experiments', query, selection)

    def fetch_trials(self, experiment=None, uid=None):
        """See :func:`~orion.storage.BaseStorageProtocol.fetch_trials`"""
        uid = get_uid(experiment, uid)

        return self._fetch_trials(dict(experiment=uid))

    def _fetch_trials(self, query, selection=None):
        """See :func:`~orion.storage.BaseStorageProtocol.fetch_trials`"""
        def sort_key(item):
            submit_time = item.submit_time
            if submit_time is None:
                return 0
            return submit_time

        trials = Trial.build(self._db.read('trials', query=query, selection=selection))
        trials.sort(key=sort_key)

        return trials

    def register_trial(self, trial):
        """See :func:`~orion.storage.BaseStorageProtocol.register_trial`"""
        self._db.write('trials', trial.to_dict())
        return trial

    def delete_trials(self, experiment=None, uid=None, where=None):
        """See :func:`~orion.storage.BaseStorageProtocol.delete_trials`"""
        uid = get_uid(experiment, uid)

        if where is None:
            where = dict()

        if uid is not None:
            where['experiment'] = uid
        return self._db.remove('trials', query=where)

    def register_lie(self, trial):
        """See :func:`~orion.storage.BaseStorageProtocol.register_lie`"""
        return self._db.write('lying_trials', trial.to_dict())

    def retrieve_result(self, trial, results_file=None, **kwargs):
        """Parse the results file that was generated by the trial process.

        Parameters
        ----------
        trial: Trial
            The trial object to be updated

        results_file: str
            the file handle to read the result from

        Returns
        -------
        returns the updated trial object

        Notes
        -----
        This does not update the database!

        """
        if results_file is None:
            return trial

        try:
            results = JSONConverter().parse(results_file.name)
        except json.decoder.JSONDecodeError:
            raise MissingResultFile()

        trial.results = [
            Trial.Result(
                name=res['name'],
                type=res['type'],
                value=res['value']) for res in results
        ]

        return trial

    def get_trial(self, trial=None, uid=None):
        """See :func:`~orion.storage.BaseStorageProtocol.get_trial`"""
        if trial is not None and uid is not None:
            assert trial._id == uid

        if uid is None:
            if trial is None:
                raise MissingArguments('Either `trial` or `uid` should be set')

            uid = trial.id

        result = self._db.read('trials', {'_id': uid})
        if not result:
            return None

        return Trial(**result[0])

    def update_trials(self, experiment=None, uid=None, where=None, **kwargs):
        """See :func:`~orion.storage.BaseStorageProtocol.update_trials`"""
        uid = get_uid(experiment, uid)
        if where is None:
            where = dict()

        where['experiment'] = uid
        return self._db.write('trials', data=kwargs, query=where)

    def update_trial(self, trial=None, uid=None, where=None, **kwargs):
        """See :func:`~orion.storage.BaseStorageProtocol.update_trial`"""
        uid = get_uid(trial, uid)

        if where is None:
            where = dict()

        where['_id'] = uid
        return self._db.write('trials', data=kwargs, query=where)

    def fetch_lost_trials(self, experiment):
        """See :func:`~orion.storage.BaseStorageProtocol.fetch_lost_trials`"""
        heartbeat = orion.core.config.worker.heartbeat
        threshold = datetime.datetime.utcnow() - datetime.timedelta(seconds=heartbeat * 5)
        lte_comparison = {'$lte': threshold}
        query = {
            'experiment': experiment._id,
            'status': 'reserved',
            'heartbeat': lte_comparison
        }

        return self._fetch_trials(query)

    def push_trial_results(self, trial):
        """See :func:`~orion.storage.BaseStorageProtocol.push_trial_results`"""
        rc = self.update_trial(trial, **trial.to_dict(),
                               where={'_id': trial.id, 'status': 'reserved'})
        if not rc:
            raise FailedUpdate()

        return rc

    def set_trial_status(self, trial, status, heartbeat=None):
        """See :func:`~orion.storage.BaseStorageProtocol.set_trial_status`"""
        if heartbeat is None:
            heartbeat = datetime.datetime.utcnow()

        update = dict(
            status=status,
            heartbeat=heartbeat,
            experiment=trial.experiment
        )

        validate_status(status)

        rc = self.update_trial(trial, **update, where={'status': trial.status, '_id': trial.id})

        if not rc:
            raise FailedUpdate()

        trial.status = status

    def fetch_pending_trials(self, experiment):
        """See :func:`~orion.storage.BaseStorageProtocol.fetch_pending_trials`"""
        query = dict(
            experiment=experiment._id,
            status={'$in': ['new', 'suspended', 'interrupted']}
        )
        return self._fetch_trials(query)

    def reserve_trial(self, experiment):
        """See :func:`~orion.storage.BaseStorageProtocol.reserve_trial`"""
        query = dict(
            experiment=experiment._id,
            status={'$in': ['interrupted', 'new', 'suspended']}
        )
        # read and write works on a single document
        now = datetime.datetime.utcnow()
        trial = self._db.read_and_write(
            'trials',
            query=query,
            data=dict(
                status='reserved',
                start_time=now,
                heartbeat=now
            ))

        if trial is None:
            return None

        return Trial(**trial)

    def fetch_noncompleted_trials(self, experiment):
        """See :func:`~orion.storage.BaseStorageProtocol.fetch_noncompleted_trials`"""
        query = dict(
            experiment=experiment._id,
            status={'$ne': 'completed'}
        )
        return self._fetch_trials(query)

    def count_completed_trials(self, experiment):
        """See :func:`~orion.storage.BaseStorageProtocol.count_completed_trials`"""
        query = dict(
            experiment=experiment._id,
            status='completed'
        )
        return self._db.count('trials', query)

    def count_broken_trials(self, experiment):
        """See :func:`~orion.storage.BaseStorageProtocol.count_broken_trials`"""
        query = dict(
            experiment=experiment._id,
            status='broken'
        )
        return self._db.count('trials', query)

    def update_heartbeat(self, trial):
        """Update trial's heartbeat"""
        return self.update_trial(trial, heartbeat=datetime.datetime.utcnow(), status='reserved')

    def fetch_trials_by_status(self, experiment, status):
        """See :func:`~orion.storage.BaseStorageProtocol.fetch_trials_by_status`"""
        query = dict(
            experiment=experiment._id,
            status=status
        )
        return self._fetch_trials(query)



"""Tests the minimalist example script on scitkit-learn and its integration to Oríon."""
import os
import subprocess

import pytest

from orion.client import create_experiment
import orion.core.cli
from orion.storage.base import get_storage


def test_script_integrity(capsys):
    """Verifies the example script can run in standalone via `python ...`."""
    script = os.path.abspath("examples/scikitlearn-iris/main.py")

    return_code = subprocess.call(["python", script, '0.1'])

    assert return_code != 2, "The example script does not exists."
    assert return_code != 1, "The example script did not terminates its execution."
    assert return_code == 0 and not capsys.readouterr().err, \
        "The example script encountered an error during its execution."


@pytest.mark.usefixtures("setup_pickleddb_database")
def test_orion_runs_script(monkeypatch):
    """Verifies Oríon can execute the example script."""
    script = os.path.abspath("examples/scikitlearn-iris/main.py")
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    config = "orion_config.yaml"

    orion.core.cli.main(["hunt", "--config", config, "python", script,
                         "orion~choices([0.1])"])

    experiment = create_experiment(name="scikit-iris-tutorial")
    assert experiment is not None
    assert experiment.version == 1

    keys = experiment.space.keys()
    assert len(keys) == 1
    assert '/_pos_2' in keys

    storage = get_storage()
    trials = storage.fetch_trials(uid=experiment.id)
    assert len(trials) == 1

    trial = trials[0]
    assert trial.status == 'completed'
    assert trial.params['/_pos_2'] == 0.1


@pytest.mark.usefixtures("setup_pickleddb_database")
def test_result_reproducibility(monkeypatch):
    """Verifies the script results stays consistent (with respect to the documentation)."""
    script = os.path.abspath("examples/scikitlearn-iris/main.py")
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    config = "orion_config.yaml"

    orion.core.cli.main(["hunt", "--config", config, "python", script,
                         "orion~choices([0.1])"])

    experiment = create_experiment(name="scikit-iris-tutorial")
    assert 'best_evaluation' in experiment.stats
    assert experiment.stats['best_evaluation'] == 0.6666666666666667



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.io.database.mongodb`."""

from datetime import datetime
import functools
from timeit import timeit

import pymongo
from pymongo import MongoClient
import pytest

from orion.core.io.database import Database, DatabaseError, DuplicateKeyError
from orion.core.io.database.mongodb import AUTH_FAILED_MESSAGES, MongoDB


@pytest.fixture(scope='module')
def orion_db():
    """Return MongoDB wrapper instance initiated with test opts."""
    MongoDB.instance = None
    orion_db = MongoDB(username='user', password='pass', name='orion_test')
    return orion_db


@pytest.fixture()
def patch_mongo_client(monkeypatch):
    """Patch ``pymongo.MongoClient`` to force serverSelectionTimeoutMS to 1."""
    def mock_class(*args, **kwargs):
        # 1 sec, defaults to 20 secs otherwise
        kwargs['serverSelectionTimeoutMS'] = 1.
        # NOTE: Can't use pymongo.MongoClient otherwise there is an infinit
        # recursion; mock(mock(mock(mock(...(MongoClient)...))))
        return MongoClient(*args, **kwargs)

    monkeypatch.setattr('pymongo.MongoClient', mock_class)


@pytest.mark.usefixtures("null_db_instances")
class TestConnection(object):
    """Create a :class:`orion.core.io.database.mongodb.MongoDB`, check connection cases."""

    @pytest.mark.usefixtures("patch_mongo_client")
    def test_bad_connection(self, monkeypatch):
        """Raise when connection cannot be achieved."""
        monkeypatch.setattr(
            MongoDB, "initiate_connection",
            MongoDB.initiate_connection.__wrapped__)
        with pytest.raises(pymongo.errors.ConnectionFailure) as exc_info:
            MongoDB(host='asdfada', port=123, name='orion',
                    username='uasdfaf', password='paasdfss')

        monkeypatch.undo()

        # Verify that the wrapper converts it properly to DatabaseError
        with pytest.raises(DatabaseError) as exc_info:
            MongoDB(host='asdfada', port=123, name='orion',
                    username='uasdfaf', password='paasdfss')
        assert "Connection" in str(exc_info.value)

    def test_bad_authentication(self, monkeypatch):
        """Raise when authentication cannot be achieved."""
        monkeypatch.setattr(
            MongoDB, "initiate_connection",
            MongoDB.initiate_connection.__wrapped__)
        with pytest.raises(pymongo.errors.OperationFailure) as exc_info:
            MongoDB(name='orion_test', username='uasdfaf', password='paasdfss')
        assert any(m in str(exc_info.value) for m in AUTH_FAILED_MESSAGES)

        monkeypatch.undo()

        with pytest.raises(DatabaseError) as exc_info:
            MongoDB(name='orion_test', username='uasdfaf', password='paasdfss')
        assert "Authentication" in str(exc_info.value)

    def test_connection_with_uri(self):
        """Check the case when connecting with ready `uri`."""
        orion_db = MongoDB('mongodb://user:pass@localhost/orion_test')
        assert orion_db.host == 'localhost'
        assert orion_db.port == 27017
        assert orion_db.username == 'user'
        assert orion_db.password == 'pass'
        assert orion_db.name == 'orion_test'

    def test_overwrite_uri(self):
        """Check the case when connecting with ready `uri`."""
        orion_db = MongoDB('mongodb://user:pass@localhost:27017/orion_test',
                           port=1231, name='orion', username='lala',
                           password='pass')
        assert orion_db.host == 'localhost'
        assert orion_db.port == 27017
        assert orion_db.username == 'user'
        assert orion_db.password == 'pass'
        assert orion_db.name == 'orion_test'

    def test_overwrite_partial_uri(self, monkeypatch):
        """Check the case when connecting with partial `uri`."""
        monkeypatch.setattr(MongoDB, 'initiate_connection', lambda self: None)

        orion_db = MongoDB('mongodb://localhost',
                           port=1231, name='orion', username='lala',
                           password='none')
        orion_db._sanitize_attrs()
        assert orion_db.host == 'localhost'
        assert orion_db.port == 1231
        assert orion_db.username == 'lala'
        assert orion_db.password == 'none'
        assert orion_db.name == 'orion'

    def test_singleton(self):
        """Test that MongoDB class is a singleton."""
        orion_db = MongoDB('mongodb://localhost',
                           port=27017, name='orion_test', username='user',
                           password='pass')
        # reinit connection does not change anything
        orion_db.initiate_connection()
        orion_db.close_connection()
        assert MongoDB() is orion_db

    def test_change_server_timeout(self):
        """Test that the server timeout is correctly changed."""
        assert timeit(lambda: MongoDB(username='user', password='pass', name='orion_test',
                                      serverSelectionTimeoutMS=1000), number=1) <= 2


@pytest.mark.usefixtures("clean_db")
class TestExceptionWrapper(object):
    """Call to methods wrapped with `mongodb_exception_wrapper()`."""

    def test_duplicate_key_error(self, monkeypatch, orion_db, exp_config):
        """Should raise generic DuplicateKeyError."""
        # Add unique indexes to force trigger of DuplicateKeyError on write()
        orion_db.ensure_index('experiments',
                              [('name', Database.ASCENDING),
                               ('metadata.user', Database.ASCENDING)],
                              unique=True)

        config_to_add = exp_config[0][0]
        config_to_add.pop('_id')

        query = {'_id': exp_config[0][1]['_id']}

        # Make sure it raises pymongo.errors.DuplicateKeyError when there is no
        # wrapper
        monkeypatch.setattr(
            orion_db, "read_and_write",
            functools.partial(orion_db.read_and_write.__wrapped__, orion_db))
        with pytest.raises(pymongo.errors.DuplicateKeyError) as exc_info:
            orion_db.read_and_write('experiments', query, config_to_add)

        monkeypatch.undo()

        # Verify that the wrapper converts it properly to DuplicateKeyError
        with pytest.raises(DuplicateKeyError) as exc_info:
            orion_db.read_and_write('experiments', query, config_to_add)
        assert "duplicate key error" in str(exc_info.value)

    def test_bulk_duplicate_key_error(self, monkeypatch, orion_db, exp_config):
        """Should raise generic DuplicateKeyError."""
        # Make sure it raises pymongo.errors.BulkWriteError when there is no
        # wrapper
        monkeypatch.setattr(
            orion_db, "write",
            functools.partial(orion_db.write.__wrapped__, orion_db))
        with pytest.raises(pymongo.errors.BulkWriteError) as exc_info:
            orion_db.write('experiments', exp_config[0])

        monkeypatch.undo()

        # Verify that the wrapper converts it properly to DuplicateKeyError
        with pytest.raises(DuplicateKeyError) as exc_info:
            orion_db.write('experiments', exp_config[0])
        assert "duplicate key error" in str(exc_info.value)

    def test_non_converted_errors(self, orion_db, exp_config):
        """Should raise OperationFailure.

        This is because _id inside exp_config[0][0] cannot be set. It is an
        immutable key of the collection.

        """
        config_to_add = exp_config[0][0]

        query = {'_id': exp_config[0][1]['_id']}

        with pytest.raises(pymongo.errors.OperationFailure):
            orion_db.read_and_write('experiments', query, config_to_add)


@pytest.mark.usefixtures("clean_db")
class TestEnsureIndex(object):
    """Calls to :meth:`orion.core.io.database.mongodb.MongoDB.ensure_index`."""

    def test_new_index(self, orion_db):
        """Index should be added to mongo database"""
        assert "status_1" not in orion_db._db.trials.index_information()
        orion_db.ensure_index('trials', 'status')
        assert "status_1" in orion_db._db.trials.index_information()

    def test_existing_index(self, orion_db):
        """Index should be added to mongo database and reattempt should do nothing"""
        assert "status_1" not in orion_db._db.trials.index_information()
        orion_db.ensure_index('trials', 'status')
        assert "status_1" in orion_db._db.trials.index_information()
        orion_db.ensure_index('trials', 'status')
        assert "status_1" in orion_db._db.trials.index_information()

    def test_ordered_index(self, orion_db):
        """Sort order should be added to index"""
        assert "end_time_-1" not in orion_db._db.trials.index_information()
        orion_db.ensure_index('trials', [('end_time', MongoDB.DESCENDING)])
        assert "end_time_-1" in orion_db._db.trials.index_information()

    def test_compound_index(self, orion_db):
        """Tuple of Index should be added as a compound index."""
        assert "name_1_metadata.user_1" not in orion_db._db.experiments.index_information()
        orion_db.ensure_index('experiments',
                              [('name', MongoDB.ASCENDING),
                               ('metadata.user', MongoDB.ASCENDING)])
        assert "name_1_metadata.user_1" in orion_db._db.experiments.index_information()

    def test_unique_index(self, orion_db):
        """Index should be set as unique in mongo database's index information."""
        assert "name_1_metadata.user_1" not in orion_db._db.experiments.index_information()
        orion_db.ensure_index('experiments',
                              [('name', MongoDB.ASCENDING),
                               ('metadata.user', MongoDB.ASCENDING)],
                              unique=True)
        index_information = orion_db._db.experiments.index_information()
        assert "name_1_metadata.user_1" in index_information
        assert index_information["name_1_metadata.user_1"]['unique']


@pytest.mark.usefixtures("clean_db")
class TestRead(object):
    """Calls to :meth:`orion.core.io.database.mongodb.MongoDB.read`."""

    def test_read_experiment(self, exp_config, orion_db):
        """Fetch a whole experiment entries."""
        loaded_config = orion_db.read(
            'trials', {'experiment': 'supernaedo2-dendi', 'status': 'new'})
        assert loaded_config == [exp_config[1][3], exp_config[1][4]]

        loaded_config = orion_db.read(
            'trials',
            {'experiment': 'supernaedo2-dendi',
             'submit_time': exp_config[1][3]['submit_time']})
        assert loaded_config == [exp_config[1][3]]
        assert loaded_config[0]['_id'] == exp_config[1][3]['_id']

    def test_read_with_id(self, exp_config, orion_db):
        """Query using ``_id`` key."""
        loaded_config = orion_db.read('experiments', {'_id': exp_config[0][2]['_id']})
        assert loaded_config == [exp_config[0][2]]

    def test_read_default(self, exp_config, orion_db):
        """Fetch value(s) from an entry."""
        value = orion_db.read(
            'experiments', {'name': 'supernaedo2', 'metadata.user': 'tsirif'},
            selection={'algorithms': 1, '_id': 0})
        assert value == [{'algorithms': exp_config[0][0]['algorithms']}]

    def test_read_nothing(self, orion_db):
        """Fetch value(s) from an entry."""
        value = orion_db.read(
            'experiments', {'name': 'not_found', 'metadata.user': 'tsirif'},
            selection={'algorithms': 1})
        assert value == []

    def test_read_trials(self, exp_config, orion_db):
        """Fetch value(s) from an entry."""
        value = orion_db.read(
            'trials',
            {'experiment': 'supernaedo2-dendi',
             'submit_time': {'$gte': datetime(2017, 11, 23, 0, 0, 0)}})
        assert value == [exp_config[1][1]] + exp_config[1][3:7]

        value = orion_db.read(
            'trials',
            {'experiment': 'supernaedo2-dendi',
             'submit_time': {'$gt': datetime(2017, 11, 23, 0, 0, 0)}})
        assert value == exp_config[1][3:7]


@pytest.mark.usefixtures("clean_db")
class TestWrite(object):
    """Calls to :meth:`orion.core.io.database.mongodb.MongoDB.write`."""

    def test_insert_one(self, database, orion_db):
        """Should insert a single new entry in the collection."""
        item = {'exp_name': 'supernaekei',
                'user': 'tsirif'}
        count_before = database.experiments.count_documents({})
        # call interface
        assert orion_db.write('experiments', item) == 1
        assert database.experiments.count_documents({}) == count_before + 1
        value = database.experiments.find_one({'exp_name': 'supernaekei'})
        assert value == item

    def test_insert_many(self, database, orion_db):
        """Should insert two new entry (as a list) in the collection."""
        item = [{'exp_name': 'supernaekei2',
                 'user': 'tsirif'},
                {'exp_name': 'supernaekei3',
                 'user': 'tsirif'}]
        count_before = database.experiments.count_documents({})
        # call interface
        assert orion_db.write('experiments', item) == 2
        assert database.experiments.count_documents({}) == count_before + 2
        value = database.experiments.find_one({'exp_name': 'supernaekei2'})
        assert value == item[0]
        value = database.experiments.find_one({'exp_name': 'supernaekei3'})
        assert value == item[1]

    def test_update_many_default(self, database, orion_db):
        """Should match existing entries, and update some of their keys."""
        filt = {'metadata.user': 'dendi'}
        count_before = database.experiments.count_documents({})
        count_filt = database.experiments.count_documents(filt)
        # call interface
        assert orion_db.write('experiments', {'pool_size': 16}, filt) == count_filt
        assert database.experiments.count_documents({}) == count_before
        value = list(database.experiments.find({}))
        assert value[0]['pool_size'] == 16
        assert value[1]['pool_size'] == 2
        assert value[2]['pool_size'] == 2
        assert value[3]['pool_size'] == 2

    def test_update_with_id(self, exp_config, database, orion_db):
        """Query using ``_id`` key."""
        filt = {'_id': exp_config[0][1]['_id']}
        count_before = database.experiments.count_documents({})
        # call interface
        assert orion_db.write('experiments', {'pool_size': 36}, filt) == 1
        assert database.experiments.count_documents({}) == count_before
        value = list(database.experiments.find())
        assert value[0]['pool_size'] == 2
        assert value[1]['pool_size'] == 36
        assert value[2]['pool_size'] == 2

    def test_no_upsert(self, database, orion_db):
        """Query with a non-existent ``_id`` should no upsert something."""
        assert orion_db.write('experiments', {'pool_size': 66}, {'_id': 'lalalathisisnew'}) == 0


@pytest.mark.usefixtures("clean_db")
class TestReadAndWrite(object):
    """Calls to :meth:`orion.core.io.database.mongodb.MongoDB.read_and_write`."""

    def test_read_and_write_one(self, database, orion_db, exp_config):
        """Should read and update a single entry in the collection."""
        # Make sure there is only one match
        documents = orion_db.read(
            'experiments',
            {'name': 'supernaedo4'})
        assert len(documents) == 1

        # Find and update atomically
        loaded_config = orion_db.read_and_write(
            'experiments',
            {'name': 'supernaedo4'},
            {'pool_size': 'lalala'})
        exp_config[0][3]['pool_size'] = 'lalala'
        assert loaded_config == exp_config[0][3]

    def test_read_and_write_many(self, database, orion_db, exp_config):
        """Should update only one entry."""
        # Make sure there is many matches
        documents = orion_db.read('experiments', {'metadata.user': 'tsirif'})
        assert len(documents) > 1

        # Find many and update first one only
        loaded_config = orion_db.read_and_write(
            'experiments',
            {'metadata.user': 'tsirif'},
            {'pool_size': 'lalala'})

        exp_config[0][1]['pool_size'] = 'lalala'
        assert loaded_config == exp_config[0][1]

        # Make sure it only changed the first document found
        documents = orion_db.read('experiments', {'metadata.user': 'tsirif'})
        assert documents[0]['pool_size'] == 'lalala'
        assert documents[1]['pool_size'] != 'lalala'

    def test_read_and_write_no_match(self, database, orion_db):
        """Should return None when there is no match."""
        loaded_config = orion_db.read_and_write(
            'experiments',
            {'name': 'lalala'},
            {'pool_size': 'lalala'})

        assert loaded_config is None


@pytest.mark.usefixtures("clean_db")
class TestRemove(object):
    """Calls to :meth:`orion.core.io.database.mongodb.MongoDB.remove`."""

    def test_remove_many_default(self, exp_config, database, orion_db):
        """Should match existing entries, and delete them all."""
        filt = {'metadata.user': 'tsirif'}
        count_before = database.experiments.count_documents({})
        count_filt = database.experiments.count_documents(filt)
        # call interface
        assert orion_db.remove('experiments', filt) == count_filt
        assert database.experiments.count_documents({}) == count_before - count_filt
        assert database.experiments.count_documents({}) == 1
        assert list(database.experiments.find()) == [exp_config[0][0]]

    def test_remove_with_id(self, exp_config, database, orion_db):
        """Query using ``_id`` key."""
        filt = {'_id': exp_config[0][0]['_id']}
        count_before = database.experiments.count_documents({})
        # call interface
        assert orion_db.remove('experiments', filt) == 1
        assert database.experiments.count_documents({}) == count_before - 1
        assert list(database.experiments.find()) == exp_config[0][1:]


@pytest.mark.usefixtures("clean_db")
class TestCount(object):
    """Calls :meth:`orion.core.io.database.mongodb.MongoDB.count`."""

    def test_count_default(self, exp_config, orion_db):
        """Call just with collection name."""
        found = orion_db.count('trials')
        assert found == len(exp_config[1])

    def test_count_query(self, exp_config, orion_db):
        """Call with a query."""
        found = orion_db.count('trials', {'status': 'completed'})
        assert found == len([x for x in exp_config[1] if x['status'] == 'completed'])

    def test_count_query_with_id(self, exp_config, orion_db):
        """Call querying with unique _id."""
        found = orion_db.count('trials', {'_id': exp_config[1][2]['_id']})
        assert found == 1

    def test_count_nothing(self, orion_db):
        """Call with argument that will not find anything."""
        found = orion_db.count('experiments', {'name': 'lalalanotfound'})
        assert found == 0


@pytest.mark.usefixtures("clean_db")
class TestIndexInformation(object):
    """Calls :meth:`orion.core.io.database.mongodb.MongoDB.count`."""

    def test_no_index(self, orion_db):
        """Test that no index is returned when there is none."""
        assert orion_db.index_information('experiments') == {'_id_': True}

    def test_single_index(self, orion_db):
        """Test with single indexes."""
        orion_db.ensure_index('experiments', [('name', MongoDB.ASCENDING)])

        assert orion_db.index_information('experiments') == {'_id_': True, 'name_1': False}

    def test_ordered_index(self, orion_db):
        """Test with ordered indexes."""
        orion_db.ensure_index('experiments', [('name', MongoDB.DESCENDING)])

        assert orion_db.index_information('experiments') == {'_id_': True, 'name_-1': False}

    def test_compound_index(self, orion_db):
        """Test representation of compound indexes."""
        orion_db.ensure_index('experiments',
                              [('name', MongoDB.DESCENDING),
                               ('version', MongoDB.ASCENDING)])

        index_info = orion_db.index_information('experiments')
        assert index_info == {'_id_': True, 'name_-1_version_1': False}

    def test_unique_index(self, orion_db):
        """Test that unique indexes are correctly inditified."""
        orion_db.ensure_index('hello',
                              [('bonjour', MongoDB.DESCENDING)], unique=True)

        assert orion_db.index_information('hello') == {'_id_': True, 'bonjour_-1': True}


@pytest.mark.usefixtures("clean_db")
class TestDropIndex(object):
    """Calls :meth:`orion.core.io.database.mongodb.MongoDB.count`."""

    def test_no_index(self, orion_db):
        """Test DatabaseError is raised when index does not exist."""
        with pytest.raises(DatabaseError) as exc:
            orion_db.drop_index('experiments', 'i_dont_exist')
        assert 'index not found with name' in str(exc.value)

    def test_drop_single_index(self, orion_db):
        """Test with single indexes."""
        orion_db.ensure_index('experiments', [('name', MongoDB.ASCENDING)])
        assert orion_db.index_information('experiments') == {'_id_': True, 'name_1': False}
        orion_db.drop_index('experiments', 'name_1')
        assert orion_db.index_information('experiments') == {'_id_': True}

    def test_drop_ordered_single_index(self, orion_db):
        """Test with single indexes."""
        orion_db.ensure_index('experiments', [('name', MongoDB.ASCENDING)])
        orion_db.ensure_index('experiments', [('name', MongoDB.DESCENDING)])
        index_info = orion_db.index_information('experiments')
        assert index_info == {'_id_': True, 'name_1': False, 'name_-1': False}
        orion_db.drop_index('experiments', 'name_1')
        index_info = orion_db.index_information('experiments')
        assert index_info == {'_id_': True, 'name_-1': False}
        with pytest.raises(DatabaseError) as exc:
            orion_db.drop_index('experiments', 'name_1')
        assert 'index not found with name' in str(exc.value)
        orion_db.drop_index('experiments', 'name_-1')
        index_info = orion_db.index_information('experiments')
        assert index_info == {'_id_': True}

    def test_drop_ordered_compound_index(self, orion_db):
        """Test with single indexes."""
        orion_db.ensure_index('experiments',
                              [('name', MongoDB.ASCENDING), ('version', MongoDB.DESCENDING)])
        orion_db.ensure_index('experiments',
                              [('name', MongoDB.DESCENDING), ('version', MongoDB.ASCENDING)])
        index_info = orion_db.index_information('experiments')
        assert index_info == {'_id_': True, 'name_1_version_-1': False, 'name_-1_version_1': False}
        orion_db.drop_index('experiments', 'name_1_version_-1')
        index_info = orion_db.index_information('experiments')
        assert index_info == {'_id_': True, 'name_-1_version_1': False}
        with pytest.raises(DatabaseError) as exc:
            orion_db.drop_index('experiments', 'name_1_version_-1')
        assert 'index not found with name' in str(exc.value)
        orion_db.drop_index('experiments', 'name_-1_version_1')
        index_info = orion_db.index_information('experiments')
        assert index_info == {'_id_': True}

    def test_drop_unique_index(self, orion_db):
        """Test with single indexes."""
        orion_db.ensure_index('hello',
                              [('bonjour', MongoDB.DESCENDING)], unique=True)
        index_info = orion_db.index_information('hello')
        assert index_info == {'_id_': True, 'bonjour_-1': True}
        orion_db.drop_index('hello', 'bonjour_-1')
        index_info = orion_db.index_information('hello')
        assert index_info == {'_id_': True}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Simple script simulating behaviour of a program that requires inputs
to determine and interact with a unique directory in local disk.
"""
import argparse
import os

from orion.client import report_results


def function(x):
    """Evaluate partial information of a quadratic."""
    z = x - 34.56789168765984988448213179176
    return 4 * z**2 + 23.4, 8 * z


def execute():
    """Execute a simple pipeline as an example."""
    parser = argparse.ArgumentParser()
    parser.add_argument('-x', type=float, required=True)
    parser.add_argument('--dir', type=str, required=True)
    parser.add_argument('--name', type=str, required=True)
    parser.add_argument('--other-name', type=str, required=True)
    inputs = parser.parse_args()

    # That's what is expected to happen
    os.makedirs(os.path.join(inputs.dir, inputs.other_name, "my-exp-{}".format(inputs.name)),
                exist_ok=False)  # Raise OSError if it exists

    y, dy = function(inputs.x)

    results = list()
    results.append(dict(
        name='example_objective',
        type='objective',
        value=y))
    results.append(dict(
        name='example_gradient',
        type='gradient',
        value=[dy]))

    report_results(results)


if __name__ == "__main__":
    execute()



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.io.database.pickleddb` -- Pickled Database
===========================================================

.. module:: database
   :platform: Unix
   :synopsis: Implement permanent version of :class:`orion.core.io.database.EphemeralDB`

"""

from contextlib import contextmanager
import logging
import os
import pickle
from pickle import PicklingError

from filelock import FileLock, Timeout

import orion.core
from orion.core.io.database import AbstractDB, DatabaseTimeout
from orion.core.io.database.ephemeraldb import EphemeralDB

log = logging.getLogger(__name__)

DEFAULT_HOST = os.path.join(orion.core.DIRS.user_data_dir, 'orion', 'orion_db.pkl')

TIMEOUT_ERROR_MESSAGE = """\
Could not acquire lock for PickledDB after {} seconds.

This is likely due to one or many of the following scenarios:

1. There is a large amount of workers and many simultaneous queries. This typically occurs
   when the task to optimize is short (few minutes). Try to reduce the amount of workers
   at least below 50.

2. The database is growing large with thousands of trials and many experiments.
   If so, you can use a different PickleDB (different file, that is, different `host`)
   for each experiment seperately to alleviate this issue.

3. The filesystem is slow. Parallel filesystems on HPC often suffer from
   large pool of users generating frequent I/O. In this case try using a separate
   partition that may be less affected.

If you cannot solve the issues listed above that are causing timeouts, you
may need to setup the MongoDB backend for better performance.
See https://orion.readthedocs.io/en/stable/install/database.html
"""


def find_unpickable_doc(dict_of_dict):
    """Look for a dictionary that cannot be pickled."""
    for name, collection in dict_of_dict.items():
        documents = collection.find()

        for doc in documents:
            try:
                pickle.dumps(doc)

            except (PicklingError, AttributeError):
                return name, doc

    return None, None


def find_unpickable_field(doc):
    """Look for a field in a dictionary that cannot be pickled"""
    if not isinstance(doc, dict):
        doc = doc.to_dict()

    for k, v in doc.items():
        try:
            pickle.dumps(v)

        except (PicklingError, AttributeError):
            return k, v

    return None, None


# pylint: disable=too-many-public-methods
class PickledDB(AbstractDB):
    """Pickled EphemeralDB to support permanancy and concurrency

    This is a very simple and inefficient implementation of a permanent database on disk for Oríon.
    The data is loaded from disk for every operation, and every operation is protected with a
    filelock.

    Parameters
    ----------
    host: str
        File path to save pickled ephemeraldb.  Default is {user data dir}/orion/orion_db.pkl ex:
        $HOME/.local/share/orion/orion_db.pkl
    timeout: int
        Maximum number of seconds to wait for the lock before raising DatabaseTimeout.
        Default is 60.

    """

    # pylint: disable=unused-argument
    def __init__(self, host=DEFAULT_HOST, timeout=60, *args, **kwargs):
        super(PickledDB, self).__init__(host)

        self.timeout = timeout

        if os.path.dirname(host):
            os.makedirs(os.path.dirname(host), exist_ok=True)

    @property
    def is_connected(self):
        """Return true, always."""
        return True

    def initiate_connection(self):
        """Do nothing"""
        pass

    def close_connection(self):
        """Do nothing"""
        pass

    def ensure_index(self, collection_name, keys, unique=False):
        """Create given indexes if they do not already exist in database.

        Indexes are only created if `unique` is True.
        """
        with self.locked_database() as database:
            database.ensure_index(collection_name, keys, unique=unique)

    def index_information(self, collection_name):
        """Return dict of names and sorting order of indexes"""
        with self.locked_database(write=False) as database:
            return database.index_information(collection_name)

    def drop_index(self, collection_name, name):
        """Remove index from the database"""
        with self.locked_database() as database:
            return database.drop_index(collection_name, name)

    def write(self, collection_name, data, query=None):
        """Write new information to a collection. Perform insert or update.

        .. seealso:: :meth:`AbstractDB.write` for argument documentation.

        """
        with self.locked_database() as database:
            return database.write(collection_name, data, query=query)

    def read(self, collection_name, query=None, selection=None):
        """Read a collection and return a value according to the query.

        .. seealso:: :meth:`AbstractDB.read` for argument documentation.

        """
        with self.locked_database(write=False) as database:
            return database.read(collection_name, query=query, selection=selection)

    def read_and_write(self, collection_name, query, data, selection=None):
        """Read a collection's document and update the found document.

        Returns the updated document, or None if nothing found.

        .. seealso:: :meth:`AbstractDB.read_and_write` for
                     argument documentation.

        """
        with self.locked_database() as database:
            return database.read_and_write(collection_name, query=query, data=data,
                                           selection=selection)

    def count(self, collection_name, query=None):
        """Count the number of documents in a collection which match the `query`.

        .. seealso:: :meth:`AbstractDB.count` for argument documentation.

        """
        with self.locked_database(write=False) as database:
            return database.count(collection_name, query=query)

    def remove(self, collection_name, query):
        """Delete from a collection document[s] which match the `query`.

        .. seealso:: :meth:`AbstractDB.remove` for argument documentation.

        """
        with self.locked_database() as database:
            return database.remove(collection_name, query=query)

    def _get_database(self):
        """Read fresh DB state from pickled file"""
        if not os.path.exists(self.host):
            return EphemeralDB()

        with open(self.host, 'rb') as f:
            data = f.read()
            if not data:
                database = EphemeralDB()
            else:
                database = pickle.loads(data)

        return database

    def _dump_database(self, database):
        """Write pickled DB on disk"""
        tmp_file = self.host + '.tmp'

        try:
            with open(tmp_file, 'wb') as f:
                pickle.dump(database, f)

        except (PicklingError, AttributeError):
            collection, doc = find_unpickable_doc(database._db)  # pylint: disable=protected-access
            log.error('Document in (collection: %s) is not pickable\ndoc: %s',
                      collection, doc.to_dict())

            key, value = find_unpickable_field(doc)
            log.error('because (value %s) in (field: %s) is not pickable',
                      value, key)
            raise

        os.rename(tmp_file, self.host)

    @contextmanager
    def locked_database(self, write=True):
        """Lock database file during wrapped operation call."""
        lock = FileLock(self.host + '.lock')

        try:
            with lock.acquire(timeout=self.timeout):
                database = self._get_database()

                yield database

                if write:
                    self._dump_database(database)
        except Timeout as e:
            raise DatabaseTimeout(TIMEOUT_ERROR_MESSAGE.format(self.timeout)) from e



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.utils.points` -- Utility functions for manipulating trial points
=================================================================================

.. module:: points
   :platform: Unix
   :synopsis: Conversion functions between higher shape points and lists.

"""


def regroup_dims(point, space):
    """Take a list of items representing a point and regroup them appropriately as
    a point from `space`.

    Parameters
    ----------
    point: array
        Points to be regrouped.
    space: `orion.algo.space.Space`
        The optimization space.

    Returns
    -------
    list or tuple

    """
    regrouped = []
    idx = 0

    for dimension in space.values():
        shape = dimension.shape
        if shape:
            assert len(shape) == 1
            next_dim = idx + shape[0]
            regrouped.append(tuple(point[idx:next_dim]))
            idx = next_dim
        else:
            regrouped.append(point[idx])
            idx += 1

    if regrouped not in space:
        raise AttributeError("The point {} is not a valid point of space {}".format(point, space))

    return regrouped


def flatten_dims(point, space):
    """Flatten `point` in `space` and convert it to a list.

    Parameters
    ----------
    point: array
        Points to be regrouped.
    space: `orion.algo.space.Space`
        The optimization space.

    Returns
    -------
    list

    """
    flattened = []

    for subpoint, dimension in zip(point, space.values()):
        shape = dimension.shape
        if shape:
            assert len(shape) == 1
            flattened.extend(subpoint)
        else:
            flattened.append(subpoint)

    return flattened



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a stress tests on python API."""
from multiprocessing import Pool
import os
import random
import time
import traceback

import matplotlib.pyplot as plt
from pymongo import MongoClient

from orion.client import create_experiment
from orion.core.io.database import DatabaseTimeout
from orion.core.utils.exceptions import SampleTimeout
from orion.core.utils.singleton import update_singletons

DB_FILE = 'stress.pkl'


def f(x, worker):
    """Sleep and return objective equal to param"""
    print(f'{worker: 6d}   {x: 5f}')
    time.sleep(max(0, random.gauss(1, 0.2)))
    return [dict(name='objective', value=x, type='objective')]


def get_experiment(storage, space_type, size):
    """Create an experiment or load from DB if already existing

    Parameters
    ----------
    storage: str
        Can be `pickleddb` or `mongodb`. A default configuration is used for each.
    space_type: str
        Can be one of
        - `discrete` Search space is discrete and limited to `max_trials`
        - `real-seeded` Search space is continuous and algos is seeded, leading to many race
          conditions while algos are sampling the same points in parallel, or
        - `real` Search space is real and algo is not seeded, leading to very few race conditions.
    size: int
        This defines `max_trials`, and the size of the search space (`uniform(0, size)`).

    """
    if storage == 'pickleddb':
        storage_config = {
            'type': 'pickleddb',
            'host': DB_FILE
        }
    elif storage == 'mongodb':
        storage_config = {
            'type': 'mongodb',
            'name': 'stress',
            'host': 'mongodb://user:pass@localhost'
        }

    discrete = space_type == 'discrete'
    high = size  # * 2

    return create_experiment(
        'stress-test',
        space={'x': f'uniform(0, {high}, discrete={discrete})'},
        max_trials=size,
        max_idle_time=60 * 5,
        algorithms={
            'random': {
                'seed': None if space_type == 'real' else 1
            }
        },
        storage={
            'type': 'legacy',
            'database': storage_config
        })


def worker(worker_id, storage, space_type, size):
    """Run trials until experiment is done

    Parameters
    ----------
    worker_id: int
        ID of the worker. This is used to distinguish logs from different workers.
    storage: str
        See `get_experiment`.
    space_type: str
        See `get_experiment`.
    size: int
        See `get_experiment`.

    """
    try:
        experiment = get_experiment(storage, space_type, size)

        assert experiment.version == 1, experiment.version

        print(f'{worker_id: 6d} enters')

        num_trials = 0
        while not experiment.is_done:
            try:
                trial = experiment.suggest()
            except SampleTimeout:
                trial = None

            if trial is None:
                break

            results = f(trial.params['x'], worker_id)
            num_trials += 1
            experiment.observe(trial, results=results)

        print(f'{worker_id: 6d} leaves | is done? {experiment.is_done}')
    except DatabaseTimeout as e:
        print(f'{worker_id: 6d} timeouts and leaves')
        return num_trials
    except Exception as e:
        print(f'{worker_id: 6d} crashes')
        traceback.print_exc()
        return None

    return num_trials


def stress_test(storage, space_type, workers, size):
    """Spawn workers and run stress test with verifications

    Parameters
    ----------
    storage: str
        See `get_experiment`.
    space_type: str
        See `get_experiment`.
    workers: int
        Number of workers to run in parallel.
    size: int
        See `get_experiment`.

    Returns
    -------
    `list` of `orion.core.worker.trial.Trial`
        List of all trials at the end of the stress test

    """
    if storage == 'pickleddb':
        if os.path.exists(DB_FILE):
            os.remove(DB_FILE)
    elif storage == 'mongodb':
        client = MongoClient(username='user', password='pass', authSource='stress')
        database = client.stress
        database.experiments.drop()
        database.lying_trials.drop()
        database.trials.drop()
        database.workers.drop()
        database.resources.drop()
        client.close()
    update_singletons()

    print('Worker  |  Point')

    with Pool(workers) as p:
        results = p.starmap(
            worker,
            zip(range(workers),
                [storage] * workers,
                [space_type] * workers,
                [size] * workers))

    assert None not in results, 'A worker crashed unexpectedly. See logs for the error messages.'
    assert all(n > 0 for n in results), 'A worker could not execute any trial.'

    if space_type in ['discrete', 'real-seeded']:
        assert sum(results) == size, results
    else:
        assert sum(results) >= size, results

    experiment = get_experiment(storage, space_type, size)

    trials = experiment.fetch_trials()

    if storage == 'pickleddb':
        os.remove(DB_FILE)
    elif storage == 'mongodb':
        client = MongoClient(username='user', password='pass', authSource='stress')
        database = client.stress
        database.experiments.drop()
        database.lying_trials.drop()
        database.trials.drop()
        database.workers.drop()
        database.resources.drop()
        client.close()
    update_singletons()

    return trials


def get_timestamps(trials, size, space_type):
    """Get start timestamps of the trials

    Parameters
    ----------
    trials: `list` of `orion.core.worker.trial.Trial`
        List of all trials at the end of the stress test
    space_type: str
        See `get_experiment`.
    size: int
        See `get_experiment`.

    Returns
    -------
    (`list`, `list`)
        Where rval[0] is start timestamp and rval[1] is the index of the trial.
        For instance the i-th trial timestamp is rval[0][rval[1].index(i)].

    """
    hparams = set()
    x = []
    y = []

    start_time = None
    for i, trial in enumerate(trials):
        hparams.add(trial.params['x'])
        assert trial.objective.value == trial.params['x']
        if start_time is None:
            start_time = trial.submit_time
        x.append((trial.submit_time - start_time).total_seconds())
        y.append(i)

    if space_type in ['discrete', 'real-seeded']:
        assert len(hparams) == size
    else:
        assert len(hparams) >= size

    return x[:size], y[:size]


def benchmark(workers, size):
    """Get start timestamps of the trials

    Parameters
    ----------
    workers: int
        see: `stress_test`.
    size: int
        See `get_experiment`.

    Returns
    -------
    dict
        Dictionary containing all results of all stress tests.
        Each key is (backend, space_type). See `get_experiment` for the supported types
        of `backend`s and `space_type`s. Each values result[(backend, space_type)] is
        in the form of a (x, y) tuple, where x a the list start timestamps and y is the indexes of
        the trials. See `get_timestamps` for more details.

    """
    results = {}
    for backend in ['mongodb', 'pickleddb']:
        for space_type in ['discrete', 'real', 'real-seeded']:
            trials = stress_test(backend, space_type, workers, size)
            results[(backend, space_type)] = get_timestamps(trials, size, space_type)

    return results


def main():
    """Run all stress tests and render the plot"""
    size = 500

    num_workers = [1, 4, 16, 32, 64, 128]

    fig, axis = plt.subplots(
        len(num_workers), 1, figsize=(5, 1.8 * len(num_workers)),
        gridspec_kw={'hspace': 0.01, 'wspace': 0},
        sharex='col')

    results = {}

    for i, workers in enumerate(num_workers):

        results[workers] = benchmark(workers, size)

        for backend in ['mongodb', 'pickleddb']:
            for space_type in ['discrete', 'real', 'real-seeded']:
                x, y = results[workers][(backend, space_type)]
                axis[i].plot(x, y, label=f'{backend}-{space_type}')

    for i, workers in enumerate(num_workers):
        # We pick 'pickleddb' and discrete=True as the reference for the slowest ones
        x, y = results[min(num_workers)][('pickleddb', 'discrete')]
        d_x = max(x) - min(x)
        d_y = max(y) - min(y)
        if i < len(num_workers) - 1:
            axis[i].text(min(x) + d_x * 0.6, min(y) + d_y * 0.1, f'{workers: 3d} workers')
        else:
            axis[i].text(min(x) + d_x * 0.6, min(y) + d_y * 0.7, f'{workers: 3d} workers')

    for i in range(len(num_workers) - 1):
        axis[i].spines['top'].set_visible(False)
        axis[i].spines['right'].set_visible(False)

    axis[-1].spines['right'].set_visible(False)
    axis[-1].spines['top'].set_visible(False)

    axis[-1].set_xlabel('Time (s)')
    axis[-1].set_ylabel('Number of trials')
    axis[-1].legend()

    plt.subplots_adjust(left=0.15, bottom=0.05, top=1, right=1)

    plt.savefig('test.png')


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# flake8: noqa: D102
# pylint: disable=no-self-use
"""
:mod:`orion.core.worker.transformer` -- Perform transformations on Dimensions
=============================================================================

.. module:: transformer
   :platform: Unix
   :synopsis: Provide functions and classes to build a Space which an
      algorithm can operate on.

"""
from abc import (ABCMeta, abstractmethod)

import numpy

from orion.algo.space import (Dimension, Space)


# pylint: disable=too-many-branches
def build_required_space(requirements, original_space):
    """Build a `Space` object which agrees to the `requirements` imposed
    by the desired optimization algorithm.

    It uses appropriate cascade of `Transformer` objects per `Dimension`
    contained in `original_space`.

    Parameters
    ----------
    requirements : list of str or str
       Describes requirements that an algorithm needs a parameter space to have
       in order to be able to operate on it. In case it is a list, the infered
       transformations are going to be applied from the first item in the list to
       the last.
    original_space : `orion.algo.space.Space`
       Original problem's definition of parameter space given by the user to Oríon.

    Supported Requirements
    ----------------------
     * Null requirement; use problem's parameter space as it is defined
     * ``'real'``: transform every dimension to a `orion.algo.space.Real` one
     * ``'integer'``: transform every dimension to a `orion.algo.space.Integer` one

    """
    requirements = requirements if isinstance(requirements, list) else [requirements]
    if not requirements:
        requirements = [None]

    space = TransformedSpace()
    for dim in original_space.values():
        transformers = []
        type_ = dim.type
        base_domain_type = type_
        for requirement in requirements:
            if type_ == 'real' and requirement in ('real', None) and dim.precision is not None:
                transformers.append(Precision(dim.precision))
            elif type_ == 'real' and requirement in ('real', None):
                pass
            elif type_ == 'real' and requirement == 'integer':
                transformers.append(Quantize())
            elif type_ == 'integer' and requirement in ('integer', None):
                pass
            elif type_ == 'integer' and requirement == 'real':
                transformers.append(Reverse(Quantize()))
            elif type_ == 'categorical' and requirement == 'real':
                transformers.extend([Enumerate(dim.categories),
                                     OneHotEncode(len(dim.categories))])
            elif type_ == 'categorical' and requirement == 'integer':
                transformers.append(Enumerate(dim.categories))
            elif type_ == 'categorical' and requirement is None:
                pass
            elif type_ == 'fidelity' and requirement is None:
                pass
            else:
                raise TypeError("Unsupported dimension type ('{}') "
                                "or requirement ('{}')".format(type_, requirement))
            try:
                last_type = transformers[-1].target_type
                type_ = last_type if last_type != 'invariant' else type_
            except IndexError:
                pass
        space.register(TransformedDimension(Compose(transformers, base_domain_type),
                                            dim))
    return space


class Transformer(object, metaclass=ABCMeta):
    """Define an (injective) function and its inverse. Base transformation class.

    :attr:`target_type` defines the type of the target space of the forward function.
    It can provide one of the values: ``['real', 'integer', 'categorical']``.

    :attr:`domain_type` is similar to `target_type` but it refers to the domain.
    If it is ``None``, then it can receive inputs of any type.
    """

    domain_type = None
    target_type = None

    @abstractmethod
    def transform(self, point):
        """Transform a point from domain dimension to the target dimension."""
        pass

    @abstractmethod
    def reverse(self, transformed_point):
        """Reverse transform a point from target dimension to the domain dimension."""
        pass

    def infer_target_shape(self, shape):
        """Return the shape of the dimension after transformation."""
        return shape

    def repr_format(self, what):
        """Format a string for calling ``__repr__`` in `TransformedDimension`."""
        return "{}({})".format(self.__class__.__name__, what)

    def _get_hashable_members(self):
        print((self.__class__.__name__, self.domain_type, self.target_type))
        return (self.__class__.__name__, self.domain_type, self.target_type)

    # pylint:disable=protected-access
    def __eq__(self, other):
        """Return True if other is the same transformed dimension as self"""
        if not isinstance(other, Transformer):
            return False
        return self._get_hashable_members() == other._get_hashable_members()


class Identity(Transformer):
    """Implement an identity transformation. Everything as it is."""

    def __init__(self, domain_type=None):
        self._domain_type = domain_type

    def transform(self, point):
        """Return `point` as it is."""
        return point

    def reverse(self, transformed_point):
        """Return `transformed_point` as it is."""
        return transformed_point

    def repr_format(self, what):
        """Format a string for calling ``__repr__`` in `TransformedDimension`."""
        return what

    @property
    def domain_type(self):
        """Return declared domain type on initialization."""
        return self._domain_type

    @property
    def target_type(self):
        """Return domain type as this will be the target in a identity transformation."""
        return self.domain_type


class Compose(Transformer):
    """Initialize composite transformer with a list of `Transformer` objects
    and domain type on which it will be applied."""

    def __init__(self, transformers, base_domain_type=None):
        try:
            self.apply = transformers.pop()
        except IndexError:
            self.apply = Identity()
        if transformers:
            self.composition = Compose(transformers, base_domain_type)
        else:
            self.composition = Identity(base_domain_type)
        assert self.apply.domain_type is None or \
            self.composition.target_type == self.apply.domain_type

    def transform(self, point):
        """Apply transformers in the increasing order of the `transformers` list."""
        point = self.composition.transform(point)
        return self.apply.transform(point)

    def reverse(self, transformed_point):
        """Reverse transformation by reversing in the opposite order of the `transformers` list."""
        transformed_point = self.apply.reverse(transformed_point)
        return self.composition.reverse(transformed_point)

    def interval(self, alpha=1.0):
        """Return interval of composed transformation."""
        if hasattr(self.apply, 'interval'):
            return self.apply.interval(alpha)

        return None

    def infer_target_shape(self, shape):
        """Return the shape of the dimension after transformation."""
        shape = self.composition.infer_target_shape(shape)
        return self.apply.infer_target_shape(shape)

    def repr_format(self, what):
        """Format a string for calling ``__repr__`` in `TransformedDimension`."""
        return self.apply.repr_format(self.composition.repr_format(what))

    @property
    def domain_type(self):
        """Return base domain type."""
        return self.composition.domain_type

    @property
    def target_type(self):
        """Infer type of the tranformation target."""
        type_before = self.composition.target_type
        type_after = self.apply.target_type
        return type_after if type_after else type_before

    # pylint:disable=protected-access
    def _get_hashable_members(self):
        return ((self.__class__.__name__, ) +
                self.apply._get_hashable_members() +
                self.composition._get_hashable_members())


class Reverse(Transformer):
    """Apply the reverse transformation that another one would do."""

    def __init__(self, transformer: Transformer):
        assert not isinstance(transformer, OneHotEncode), "real to categorical is pointless"
        self.transformer = transformer

    def transform(self, point):
        """Use `reserve` of composed `transformer`."""
        return self.transformer.reverse(point)

    def reverse(self, transformed_point):
        """Use `transform` of composed `transformer`."""
        return self.transformer.transform(transformed_point)

    def repr_format(self, what):
        """Format a string for calling ``__repr__`` in `TransformedDimension`."""
        return "{}{}".format(self.__class__.__name__, self.transformer.repr_format(what))

    @property
    def target_type(self):
        """Return `domain_type` of composed `transformer`."""
        return self.transformer.domain_type

    @property
    def domain_type(self):
        """Return `target_type` of composed `transformer`."""
        return self.transformer.target_type


class Precision(Transformer):
    """Round real numbers to requested precision."""

    domain_type = 'real'
    target_type = 'real'

    def __init__(self, precision=4):
        self.precision = precision

    def transform(self, point):
        """Round `point` to the requested precision, as numpy arrays."""
        # numpy.format_float_scientific precision starts at 0
        if isinstance(point, (list, tuple)) or (isinstance(point, numpy.ndarray) and point.shape):
            point = map(lambda x: numpy.format_float_scientific(x, precision=self.precision - 1),
                        point)
            point = list(map(float, point))
        else:
            point = float(numpy.format_float_scientific(point, precision=self.precision - 1))

        return numpy.asarray(point)

    def reverse(self, transformed_point):
        """Cast `transformed_point` to floats, as numpy arrays."""
        return self.transform(transformed_point)

    def repr_format(self, what):
        """Format a string for calling ``__repr__`` in `TransformedDimension`."""
        return "{}({}, {})".format(self.__class__.__name__, self.precision, what)


class Quantize(Transformer):
    """Transform real numbers to integers, violating injection."""

    domain_type = 'real'
    target_type = 'integer'

    def transform(self, point):
        """Cast `point` to the floor and then to integers, as numpy arrays."""
        return numpy.floor(numpy.asarray(point)).astype(int)

    def reverse(self, transformed_point):
        """Cast `transformed_point` to floats, as numpy arrays."""
        return numpy.asarray(transformed_point).astype(float)


class Enumerate(Transformer):
    """Enumerate categories.

    Effectively transform from a list of objects to a range of integers.
    """

    domain_type = 'categorical'
    target_type = 'integer'

    def __init__(self, categories):
        self.categories = categories
        map_dict = {cat: i for i, cat in enumerate(categories)}
        self._map = numpy.vectorize(lambda x: map_dict[x], otypes='i')
        self._imap = numpy.vectorize(lambda x: categories[x], otypes=[numpy.object])

    def __deepcopy__(self, memo):
        """Make a deepcopy"""
        return type(self)(self.categories)

    def transform(self, point):
        """Return integers corresponding uniquely to the categories in `point`.

        :rtype: numpy.ndarray, integer
        """
        return self._map(point)

    def reverse(self, transformed_point):
        """Return categories corresponding to their positions inside `transformed_point`.

        :rtype: numpy.ndarray, numpy.object
        """
        return self._imap(transformed_point)


class OneHotEncode(Transformer):
    """Encode categories to a 1-hot integer space representation."""

    domain_type = 'integer'
    target_type = 'real'

    def __init__(self, bound: int):
        self.num_cats = bound

    def transform(self, point):
        """Match a `point` containing integers to real vector representations of them.

        If the upper bound of integers supported by an instance of `OneHotEncode`
        is less or equal to 2, then cast them to floats.

        .. note:: This transformation possibly appends one more tensor dimension to `point`.
        """
        point_ = numpy.asarray(point)
        assert numpy.all(point_ < self.num_cats) and numpy.all(point_ >= 0) and\
            numpy.all(point_ % 1 == 0)

        if self.num_cats <= 2:
            return numpy.asarray(point_, dtype=float)

        hot = numpy.zeros(self.infer_target_shape(point_.shape))
        grid = numpy.meshgrid(*[numpy.arange(dim) for dim in point_.shape],
                              indexing='ij')
        hot[grid + [point_]] = 1
        return hot

    def reverse(self, transformed_point):
        """Match real vector representations to integers using an argmax function.

        If the number of dimensions is exactly 2, then use 0.5 as a decision boundary,
        and convert representation to integers 0 or 1.

        If the number of dimensions is exactly 1, then return zeros.

        .. note:: This reverse transformation possibly removes the last tensor dimension
           from `transformed_point`.
        """
        point_ = numpy.asarray(transformed_point)
        if self.num_cats == 2:
            return (point_ > 0.5).astype(int)
        elif self.num_cats == 1:
            return numpy.zeros_like(point_, dtype=int)

        assert point_.shape[-1] == self.num_cats
        return point_.argmax(axis=-1)

    # pylint:disable=unused-argument
    def interval(self, alpha=1.0):
        """Return the interval for the one-hot encoding in proper shape."""
        if self.num_cats == 2:
            return 0, 1
        else:
            low = numpy.zeros(self.num_cats)
            high = numpy.ones(self.num_cats)

            return low, high

    def infer_target_shape(self, shape):
        """Infer that transformed points will have one more tensor dimension,
        if the number of supported integers to transform is larger than 2.
        """
        return tuple(list(shape) + [self.num_cats]) if self.num_cats > 2 else shape

    def _get_hashable_members(self):
        return super(OneHotEncode, self)._get_hashable_members() + (self.num_cats, )


# pylint:disable=too-many-public-methods
class TransformedDimension(object):
    """Duck-type `Dimension` to mimic its functionality,
    while transform automatically and appropriately an underlying `Dimension` object
    according to a `Transformer` object.
    """

    NO_DEFAULT_VALUE = Dimension.NO_DEFAULT_VALUE

    def __init__(self, transformer: Transformer, original_dimension: Dimension):
        """Initialize a `TransformedDimension` with an `original_dimension` object
        and the `transformer` that will be used.
        """
        self.original_dimension = original_dimension
        self.transformer = transformer

    def transform(self, point):
        """Expose `Transformer.transform` interface from underlying instance."""
        return self.transformer.transform(point)

    def reverse(self, transformed_point):
        """Expose `Transformer.reverse` interface from underlying instance."""
        return self.transformer.reverse(transformed_point)

    def sample(self, n_samples=1, seed=None):
        """Sample from the original dimension and forward transform them."""
        samples = self.original_dimension.sample(n_samples, seed)
        return [self.transform(sample) for sample in samples]

    def interval(self, alpha=1.0):
        """Map the interval bounds to the transformed ones."""
        if hasattr(self.transformer, 'interval'):
            interval = self.transformer.interval()
            if interval:
                return interval
        elif self.original_dimension.prior_name == 'choices':
            return self.original_dimension.categories

        low, high = self.original_dimension.interval(alpha)

        return self.transform(low), self.transform(high)

    def __contains__(self, point):
        """Reverse transform and ask the original dimension if it is a possible
        sample.
        """
        try:
            orig_point = self.reverse(point)
        except AssertionError:
            return False
        return orig_point in self.original_dimension

    def __repr__(self):
        """Represent the object as a string."""
        return self.transformer.repr_format(repr(self.original_dimension))

    # pylint:disable=protected-access
    def __eq__(self, other):
        """Return True if other is the same transformed dimension as self"""
        if not (hasattr(other, "transformer") and hasattr(other, "original_dimension")):
            return False

        return (self.transformer == other.transformer and
                self.original_dimension == other.original_dimension)

    def __hash__(self):
        return hash(self._get_hashable_members())

    # pylint:disable=protected-access
    def _get_hashable_members(self):
        print(self.transformer._get_hashable_members())
        print(self.original_dimension._get_hashable_members())
        return (self.transformer._get_hashable_members() +
                self.original_dimension._get_hashable_members())

    def validate(self):
        """Validate original_dimension"""
        self.original_dimension.validate()

    def get_prior_string(self):
        """Do not change the prior string of original dimension."""
        return self.transformer.repr_format(self.original_dimension.get_prior_string())

    def get_string(self):
        """Do not change the string of original dimension."""
        original_prior = self.original_dimension.get_prior_string()
        original_string = self.original_dimension.get_string()
        return original_string.replace(original_prior, self.get_prior_string())

    @property
    def name(self):
        """Do not change the name of the original dimension."""
        return self.original_dimension.name

    @property
    def type(self):
        """Ask transformer which is its target class."""
        type_ = self.transformer.target_type
        return type_ if type_ != 'invariant' else self.original_dimension.type

    @property
    def prior_name(self):
        """Do not change the prior name of the original dimension."""
        return self.original_dimension.prior_name

    @property
    def shape(self):
        """Wrap original shape with transformer, because it may have changed."""
        return self.transformer.infer_target_shape(self.original_dimension.shape)

    @property
    def default_value(self):
        """Wrap original default value."""
        defval = self.original_dimension.default_value
        return self.transform(defval) if defval is not None else None

    def cast(self, point):
        """Cast a point according to original_dimension and then transform it"""
        return self.transform(self.original_dimension.cast(point))

    @property
    def cardinality(self):
        """Wrap original `Dimension` capacity"""
        return self.original_dimension.cardinality


class TransformedSpace(Space):
    """Wrap the `Space` to support transformation methods."""

    contains = TransformedDimension

    def transform(self, point):
        """Transform a point that was in the original space to be in this one."""
        return tuple([dim.transform(point[i]) for i, dim in enumerate(self.values())])

    def reverse(self, transformed_point):
        """Reverses transformation so that a point from this `TransformedSpace`
        to be in the original one.
        """
        return tuple([dim.reverse(transformed_point[i]) for i, dim in enumerate(self.values())])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.cli.info`."""
import itertools

import pytest

from orion.core.io.space_builder import SpaceBuilder
from orion.core.utils.format_terminal import (
    format_algorithm, format_commandline, format_config, format_dict, format_identification,
    format_info, format_list, format_metadata, format_refers, format_space, format_stats,
    format_title, get_trial_params)
from orion.core.worker.trial import Trial


class DummyExperiment():
    """Dummy container to mock experiments"""

    pass


@pytest.fixture
def dummy_trial():
    """Return a dummy trial object"""
    trial = Trial()
    trial._params = [
        Trial.Param(name='a', type='real', value=0.0),
        Trial.Param(name='b', type='integer', value=1),
        Trial.Param(name='c', type='categorical', value='Some')]
    return trial


@pytest.fixture
def dummy_dict():
    """Return a dict of dicts"""
    return {
        1: {
            1.1: "1.1.1",
            1.2: {
                "1.2.1": {},
                "1.2.2": "1.2.2.1"
            }
        },
        2: {
            2.1: "2.1.1",
            2.2: {}
        },
        3: {}
    }


@pytest.fixture
def dummy_list_of_lists():
    """Return a list of lists"""
    return [
        1,
        [2, 3],
        4,
        [5, 6, 7, 8]
    ]


@pytest.fixture
def dummy_list_of_objects(dummy_dict):
    """Return a list of objects"""
    return [
        {
            1: {
                1.1: "1.1.1",
                1.2: [
                    "1.2.1",
                    "1.2.2"
                ]
            },
        },
        [
            4, 5
        ],
        {
            2: {
                2.1: "2.1.1",
                2.2: {}
            },
            3: {}
        }
    ]


@pytest.fixture
def algorithm_dict():
    """Return an algorithm configuration"""
    return dict(
        bayesianoptimizer=dict(
            acq_func='gp_hedge',
            alpha=1e-10,
            n_initial_points=10,
            n_restarts_optimizer=0,
            normalize_y=False))


def test_format_title():
    """Test title formatting template"""
    result = """Test\n===="""

    assert format_title('Test') == result


@pytest.mark.parametrize("depth", [0, 1, 2])
def test_format_dict_depth_synthetic(depth, dummy_dict):
    """Test dict formatting with different depths for one line"""
    WIDTH = 4
    tab = (" " * WIDTH) * depth
    assert format_dict(dummy_dict, depth=depth, width=WIDTH).split("\n")[0].startswith(tab + "1")


def test_format_dict_depth_full(dummy_dict):
    """Test dict depth formatting for all lines"""
    WIDTH = 4
    tab = " " * WIDTH

    lines = format_dict(dummy_dict).split("\n")
    assert len(lines) == 9
    assert lines[0].startswith("1")
    assert lines[1].startswith(tab + "1")
    assert lines[2].startswith(tab + "1")
    assert lines[3].startswith((tab * 2) + "1")
    assert lines[4].startswith((tab * 2) + "1")
    assert lines[5].startswith("2")
    assert lines[6].startswith(tab + "2")
    assert lines[7].startswith(tab + "2")
    assert lines[8].startswith("3")


@pytest.mark.parametrize("width,depth",
                         itertools.product([0, 2], [1, 2]))
def test_format_dict_width_synthetic(width, depth, dummy_dict):
    """Test dict formatting with different combination of widths and depth for one line"""
    tab = (" " * width) * depth
    assert format_dict(dummy_dict, depth=depth, width=width).split("\n")[0].startswith(tab + "1")


@pytest.mark.parametrize("width", [0, 5, 12])
def test_format_dict_width_full(width, dummy_dict):
    """Test dict formatting with different widths for all lines"""
    tab = " " * width
    lines = format_dict(dummy_dict, width=width).split("\n")
    assert len(lines) == 9
    assert lines[0].startswith("1")
    assert lines[1].startswith(tab + "1")
    assert lines[2].startswith(tab + "1")
    assert lines[3].startswith((tab * 2) + "1")
    assert lines[4].startswith((tab * 2) + "1")
    assert lines[5].startswith("2")
    assert lines[6].startswith(tab + "2")
    assert lines[7].startswith(tab + "2")
    assert lines[8].startswith("3")


def test_format_dict_empty_leaf_template():
    """Test dict empty leaf node formatting"""
    dummy_dict_with_leafs = {1: {2: {}}, 2: {}, 3: {1: 3, 2: {}}}
    lines = format_dict(dummy_dict_with_leafs).split("\n")
    assert len(lines) == 6
    # 1:
    assert lines[1].lstrip(" ") == "2"
    assert lines[2].lstrip(" ") == "2"
    # 3:
    #   1: 3
    assert lines[5].lstrip(" ") == "2"


def test_format_dict_empty_leaf_template_custom():
    """Test dict empty leaf node formatting with custom template"""
    template = "{key} is a leaf\n"
    dummy_dict_with_leafs = {1: {2: {}}, 2: {}, 3: {1: 3, 2: {}}}
    lines = format_dict(dummy_dict_with_leafs, templates={'empty_leaf': template}).split("\n")
    assert len(lines) == 6
    # 1:
    assert lines[1] == "2 is a leaf"
    assert lines[2] == "2 is a leaf"
    # 3:
    #   1: 3
    assert lines[5] == "2 is a leaf"


def test_format_dict_leaf_template():
    """Test dict leaf node formatting"""
    dummy_dict_with_leafs = {1: {2: {}}, 2: {}, 3: {1: 3, 2: {}}}
    lines = format_dict(dummy_dict_with_leafs).split("\n")
    assert len(lines) == 6
    # 1:
    #   2
    # 2
    # 3:
    assert lines[4].lstrip(" ") == "1: 3"


def test_format_dict_leaf_template_custom():
    """Test dict leaf node formatting with custom template"""
    template = "value of {key} is {value}\n"
    dummy_dict_with_leafs = {1: {2: {}}, 2: {}, 3: {1: 3, 2: {}}}
    lines = format_dict(dummy_dict_with_leafs, templates={'leaf': template}).split("\n")
    assert len(lines) == 6
    # 1:
    #   2
    # 2
    # 3:
    assert lines[4] == "value of 1 is 3"


def test_format_dict_node_template():
    """Test dict node formatting"""
    dummy_dict_with_leafs = {1: {2: {}}, 2: {}, 3: {1: {4: 3}, 2: {}}}
    lines = format_dict(dummy_dict_with_leafs).split("\n")
    assert len(lines) == 7
    assert lines[0] == "1:"
    #   2
    # 2
    assert lines[3] == "3:"
    assert lines[4].lstrip(" ") == "1:"
    #     4: 3
    #   2


def test_format_dict_node_template_custom():
    """Test dict formatting with custom node template"""
    template = "{key} is a dict:\n{value}\n"
    dummy_dict_with_leafs = {1: {2: {}}, 2: {}, 3: {1: {4: 3}, 2: {}}}
    lines = format_dict(dummy_dict_with_leafs, templates={'dict_node': template}).split("\n")
    assert len(lines) == 7
    assert lines[0] == "1 is a dict:"
    #   2
    # 2
    assert lines[3] == "3 is a dict:"
    assert lines[4] == "1 is a dict:"
    #     4: 3
    #   2


@pytest.mark.parametrize("depth", [0, 1, 2])
def test_format_list_depth_synthetic(depth, dummy_list_of_lists):
    """Test list of lists formatting with different depths for one line"""
    WIDTH = 4
    tab = (" " * WIDTH) * depth
    formatted_list_str = format_list(dummy_list_of_lists, depth=depth, width=WIDTH)
    assert formatted_list_str.split("\n")[0].startswith(tab + "[")


def test_format_list_depth_full(dummy_list_of_lists):
    """Test list of lists depth formatting for all lines"""
    WIDTH = 4
    tab = " " * WIDTH

    lines = format_list(dummy_list_of_lists).split("\n")
    assert len(lines) == 14
    assert lines[0] == "["
    assert lines[1] == tab + "1"
    assert lines[2] == tab + "["
    assert lines[3] == (tab * 2) + "2"
    assert lines[4] == (tab * 2) + "3"
    assert lines[5] == tab + "]"
    assert lines[6] == tab + "4"
    assert lines[7] == tab + "["
    assert lines[8] == (tab * 2) + "5"
    assert lines[9] == (tab * 2) + "6"
    assert lines[10] == (tab * 2) + "7"
    assert lines[11] == (tab * 2) + "8"
    assert lines[12] == tab + "]"
    assert lines[13] == "]"


@pytest.mark.parametrize("width,depth",
                         itertools.product([0, 2], [1, 2]))
def test_format_list_width_synthetic(width, depth, dummy_list_of_lists):
    """Test list of lists formatting with different combination of widths and depth for one line"""
    tab = (" " * width) * depth
    assert format_list(dummy_list_of_lists, depth=depth, width=width).split("\n")[0] == tab + "["


@pytest.mark.parametrize("width", [0, 5, 12])
def test_format_list_width_full(width, dummy_list_of_lists):
    """Test list of lists formatting with different widths for all lines"""
    tab = " " * width
    lines = format_list(dummy_list_of_lists, width=width).split("\n")
    assert len(lines) == 14
    assert lines[0] == "["
    assert lines[1] == tab + "1"
    assert lines[2] == tab + "["
    assert lines[3] == (tab * 2) + "2"
    assert lines[4] == (tab * 2) + "3"
    assert lines[5] == tab + "]"
    assert lines[6] == tab + "4"
    assert lines[7] == tab + "["
    assert lines[8] == (tab * 2) + "5"
    assert lines[9] == (tab * 2) + "6"
    assert lines[10] == (tab * 2) + "7"
    assert lines[11] == (tab * 2) + "8"
    assert lines[12] == tab + "]"
    assert lines[13] == "]"


def test_format_list_item_template(dummy_list_of_lists):
    """Test list of lists formatting of items"""
    lines = format_list(dummy_list_of_lists).split("\n")
    assert len(lines) == 14
    assert lines[1].lstrip(" ") == "1"
    assert lines[3].lstrip(" ") == "2"


def test_format_list_item_template_custom(dummy_list_of_lists):
    """Test list of lists with custom item template"""
    template = "{item} is an item\n"
    lines = format_list(dummy_list_of_lists, templates={'item': template}).split("\n")
    assert len(lines) == 14
    # 1:
    assert lines[1] == "1 is an item"
    assert lines[3] == "2 is an item"


def test_format_list_node_template(dummy_list_of_lists):
    """Test list of lists formatting of node"""
    WIDTH = 4
    tab = " " * WIDTH
    lines = format_list(dummy_list_of_lists).split("\n")
    assert len(lines) == 14
    assert lines[2] == tab + "["
    assert lines[3] == (tab * 2) + "2"
    assert lines[4] == (tab * 2) + "3"
    assert lines[5] == tab + "]"


def test_format_list_node_template_custom(dummy_list_of_lists):
    """Test list of lists custom formatting"""
    templates = dict(
        list="[{items}]",
        item="{item}",
        list_node="{item}")
    lines = format_list(dummy_list_of_lists, templates=templates).split("\n")
    assert len(lines) == 1
    assert lines[0] == "[1[23]4[5678]]"


def test_format_dict_with_list(dummy_list_of_objects):
    """Test dict formatting with embedded lists"""
    assert format_dict(dummy_list_of_objects) == """\
[
    1:
        1.1: 1.1.1
        1.2:
            [
                1.2.1
                1.2.2
            ]
    [
        4
        5
    ]
    2:
        2.1: 2.1.1
        2.2
    3
]\
"""


def test_format_identification():
    """Test commandline section formatting"""
    experiment = DummyExperiment()
    experiment.name = "test"
    experiment.version = 1
    experiment.metadata = {'user': 'corneauf'}
    assert format_identification(experiment) == """\
Identification
==============
name: test
version: 1
user: corneauf
"""


def test_format_commandline():
    """Test commandline section formatting"""
    experiment = DummyExperiment()
    commandline = ['executing.sh', '--some', 'random', '--command', 'line', 'arguments']
    experiment.metadata = {'user_args': commandline}
    assert format_commandline(experiment) == """\
Commandline
===========
executing.sh --some random --command line arguments
"""


def test_format_config(monkeypatch):
    """Test config section formatting"""
    experiment = DummyExperiment()
    experiment.pool_size = 10
    experiment.max_trials = 100
    experiment.working_dir = "working_dir"
    assert format_config(experiment) == """\
Config
======
pool size: 10
max trials: 100
working dir: working_dir
"""


def test_format_algorithm(algorithm_dict):
    """Test algorithm section formatting"""
    experiment = DummyExperiment()
    experiment.configuration = {'algorithms': algorithm_dict}
    assert format_algorithm(experiment) == """\
Algorithm
=========
bayesianoptimizer:
    acq_func: gp_hedge
    alpha: 1e-10
    n_initial_points: 10
    n_restarts_optimizer: 0
    normalize_y: False
"""


def test_format_space():
    """Test space section formatting"""
    experiment = DummyExperiment()
    space = SpaceBuilder().build(
        {"some": 'choices(["random", "or", "not"])',
         "command": 'uniform(0, 1)'})
    experiment.space = space
    assert format_space(experiment) == """\
Space
=====
command: uniform(0, 1)
some: choices(['random', 'or', 'not'])
"""


def test_format_metadata():
    """Test metadata section formatting"""
    experiment = DummyExperiment()
    experiment.metadata = dict(
        user='user',
        datetime='now',
        orion_version='1.0.1',
        VCS=dict(
            HEAD_sha='sha',
            active_branch='branch',
            diff_sha='smt',
            is_dirty=True,
            type='git'))
    assert format_metadata(experiment) == """\
Meta-data
=========
user: user
datetime: now
orion version: 1.0.1
VCS:
  HEAD_sha: sha
  active_branch: branch
  diff_sha: smt
  is_dirty: True
  type: git
"""


def test_format_refers_root():
    """Test refers section formatting for a root experiment"""
    experiment = DummyExperiment()
    experiment.node = DummyExperiment()
    experiment.node.root = experiment.node

    # experiment.refers = dict(
    #     parent='user',
    #     datetime='now',
    #     orion_version='1.0.1')
    assert format_refers(experiment) == """\
Parent experiment
=================
root: 
parent: 
adapter: 
"""  # noqa: W291


def test_format_refers_child():
    """Test refers section formatting for a child experiment"""
    ROOT_NAME = 'root-name'
    PARENT_NAME = 'parent-name'

    root = DummyExperiment()
    root.name = ROOT_NAME

    parent = DummyExperiment()
    parent.name = PARENT_NAME

    child = DummyExperiment()
    child.node = DummyExperiment()
    child.node.parent = parent
    child.node.root = root

    adapter = DummyExperiment()
    adapter.configuration = dict(
        adummy='dict',
        foran='adapter')

    child.refers = dict(adapter=adapter)

    # experiment.refers = dict(
    #     parent='user',
    #     datetime='now',
    #     orion_version='1.0.1')
    assert format_refers(child) == """\
Parent experiment
=================
root: root-name
parent: parent-name
adapter: 
  adummy: dict
  foran: adapter
"""  # noqa: W291


def test_get_trial_params_empty():
    """Test failing to fetch trials does not fail"""
    experiment = DummyExperiment()
    experiment.get_trial = lambda trial=None, uid=None: None
    assert get_trial_params(None, experiment) == {}


def test_get_trial_params(dummy_trial):
    """Test params are converted properly to a dict."""
    experiment = DummyExperiment()
    experiment.get_trial = lambda trial=None, uid=None: dummy_trial
    params = get_trial_params(None, experiment)
    assert params['a'] == 0.0
    assert params['b'] == 1
    assert params['c'] == 'Some'


def test_format_stats(dummy_trial):
    """Test stats section formatting"""
    experiment = DummyExperiment()
    experiment.stats = dict(
        best_trials_id='dummy',
        trials_completed=10,
        best_evaluation=0.1,
        start_time='yesterday',
        finish_time='now',
        duration='way too long')
    experiment.get_trial = lambda trial=None, uid=None: dummy_trial
    experiment.is_done = False
    assert format_stats(experiment) == """\
Stats
=====
completed: False
trials completed: 10
best trial:
  id: dummy
  evaluation: 0.1
  params:
    a: 0.0
    b: 1
    c: Some
start time: yesterday
finish time: now
duration: way too long
"""


def test_format_info(algorithm_dict, dummy_trial):
    """Test full formatting string"""
    experiment = DummyExperiment()
    commandline = ['executing.sh', '--some~choices(["random", "or", "not"])',
                   '--command~uniform(0, 1)']
    experiment.name = 'test'
    experiment.version = 1
    experiment.metadata = {'user_args': commandline}
    experiment.pool_size = 10
    experiment.max_trials = 100
    experiment.working_dir = "working_dir"
    experiment.configuration = {'algorithms': algorithm_dict}

    space = SpaceBuilder().build(
        {"some": 'choices(["random", "or", "not"])',
         "command": 'uniform(0, 1)'})
    experiment.space = space
    experiment.metadata.update(dict(
        user='user',
        datetime='now',
        orion_version='1.0.1',
        VCS=dict(
            HEAD_sha='sha',
            active_branch='branch',
            diff_sha='smt',
            is_dirty=True,
            type='git')))

    ROOT_NAME = 'root-name'
    PARENT_NAME = 'parent-name'

    root = DummyExperiment()
    root.name = ROOT_NAME

    parent = DummyExperiment()
    parent.name = PARENT_NAME

    experiment.node = DummyExperiment()
    experiment.node.parent = parent
    experiment.node.root = root

    adapter = DummyExperiment()
    adapter.configuration = dict(
        adummy='dict',
        foran='adapter')

    experiment.refers = dict(adapter=adapter)
    experiment.stats = dict(
        best_trials_id='dummy',
        trials_completed=10,
        best_evaluation=0.1,
        start_time='yesterday',
        finish_time='now',
        duration='way too long')
    experiment.get_trial = lambda trial=None, uid=None: dummy_trial
    experiment.is_done = False

    assert format_info(experiment) == """\
Identification
==============
name: test
version: 1
user: user


Commandline
===========
executing.sh --some~choices(["random", "or", "not"]) --command~uniform(0, 1)


Config
======
pool size: 10
max trials: 100
working dir: working_dir


Algorithm
=========
bayesianoptimizer:
    acq_func: gp_hedge
    alpha: 1e-10
    n_initial_points: 10
    n_restarts_optimizer: 0
    normalize_y: False


Space
=====
command: uniform(0, 1)
some: choices(['random', 'or', 'not'])


Meta-data
=========
user: user
datetime: now
orion version: 1.0.1
VCS:
  HEAD_sha: sha
  active_branch: branch
  diff_sha: smt
  is_dirty: True
  type: git


Parent experiment
=================
root: root-name
parent: parent-name
adapter: 
  adummy: dict
  foran: adapter


Stats
=====
completed: False
trials completed: 10
best trial:
  id: dummy
  evaluation: 0.1
  params:
    a: 0.0
    b: 1
    c: Some
start time: yesterday
finish time: now
duration: way too long

"""  # noqa: W291



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.serve` -- Web application endpoint
=======================================================

.. module:: serve
   :platform: Unix
   :synopsis: Starts an http endpoint to serve requests

"""
import argparse
import logging

from gunicorn.app.base import BaseApplication

import orion.core.io.experiment_builder as experiment_builder
from orion.serving.webapi import WebApi


log = logging.getLogger(__name__)
DESCRIPTION = "Starts Oríon's REST API server"


def add_subparser(parser):
    """Add the subparser that needs to be used for this command"""
    serve_parser = parser.add_parser('serve', help=DESCRIPTION, description=DESCRIPTION)

    serve_parser.add_argument('-c', '--config', type=argparse.FileType('r'),
                              metavar='path-to-config', help="user provided "
                              "orion configuration file")

    serve_parser.set_defaults(func=main)

    return serve_parser


def main(args):
    """Starts an application server to serve http requests"""
    config = experiment_builder.get_cmd_config(args)

    web_api = WebApi(config)

    gunicorn_app = GunicornApp(web_api)
    gunicorn_app.run()


class GunicornApp(BaseApplication):
    """Custom Gunicorn application, required when integrating gunicorn as an API."""

    def __init__(self, app, options=None):
        self.options = options or {}
        self.application = app
        super(GunicornApp, self).__init__()

    def init(self, parser, opts, args):
        """Pre-run initialization"""
        pass

    def load_config(self):
        """Load the gunicorn config"""
        for key, value in self.options.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        """Load the WSGI application"""
        return self.application



# -*- coding: utf-8 -*-
# pylint: disable=eval-used,protected-access
"""
:mod:`orion.core.io.space_builder` -- Create Space objects from configuration
=============================================================================

.. module:: space_builder
   :platform: Unix
   :synopsis: Functions which build `Dimension` and `Space` objects for
      defining problem's search space.

Replace actual hyperparam values in your script's config files or cmd
arguments with orion's keywords for declaring hyperparameter types
to be optimized.

Motivation for this way of orion's configuration is to achieve as
minimal intrusion to user's workflow as possible by:

   * Offering to user the choice to keep the original way of passing
     hyperparameters to their script, be it through some **config file
     type** (e.g. yaml, json, ini, etc) or through **command line
     arguments**.

   * Instead of passing the actual hyperparameter values, use one of
     the characteristic keywords, names enlisted in `scipy.stats.distributions`
     or `orion.core.io.space_builder.DimensionBuilder`,
     to describe distributions and declare the hyperparameters
     to be optimized. So that a possible command line argument
     like ``-lrate0=0.1`` becomes ``-lrate0~'uniform(-3, 1)'``.

.. note::
   Use ``~`` instead of ``=`` to denote that a variable "draws from"
   a distribution. We support limited Python syntax for describing distributions.

   * Module will also use the script's provided input file/args as a
     template to fill an appropriate input with proposed values for the
     script's execution in each hyperiteration.

"""
from collections import OrderedDict
import logging
import re

from scipy.stats import distributions as sp_dists

from orion.algo.space import (Categorical, Fidelity, Integer, Real, Space)
from orion.core.utils.flatten import flatten


log = logging.getLogger(__name__)


def _check_expr_to_eval(expr):
    if '__' in expr or ';' in expr:
        raise RuntimeError("Cannot use builtins, '__' or ';'. Sorry.")
    return


def _get_arguments(*args, **kwargs):
    return args, kwargs


def _real_or_int(kwargs):
    return Integer if kwargs.pop('discrete', False) else Real


def replace_key_in_order(odict, key_prev, key_after):
    """Replace `key_prev` of `OrderedDict` `odict` with `key_after`,
    while leaving its value and the rest of the dictionary intact and in the
    same order.
    """
    tmp = OrderedDict()
    for k, v in odict.items():
        if k == key_prev:
            tmp[key_after] = v
        else:
            tmp[k] = v
    return tmp


def _should_not_be_built(expression):
    return expression.startswith('-') or expression.startswith('>')


def _remove_marker(expression, marker='+'):
    return expression.replace(marker, '', 1) if expression.startswith(marker) else expression


class DimensionBuilder(object):
    """Create `Dimension` objects using a name for it and an string expression
    which encodes prior and dimension information.

    Basically, one must provide a string like a function call to a method that
    has the name of a distribution, .e.g. ``alpha``, and then provide settings
    about that distributions and information about the `Dimension`, if it
    cannot be inferred. One example for the latter case is:

    ``uniform(-3, 5)`` will return a `Real` dimension, while
    ``uniform(-3, 5, discrete=True)`` will return an `Integer` dimension.

    Sometimes there is also a separate name for the same distribution in integers,
    for the 'uniform' example:

    ``randint(-3, 5)`` will return a uniform `Integer` dimension.

    For categorical dimensions, one can use either ``enum`` or ``random`` name.
    ``random`` however can be used for uniform reals or integers as well.

    Most names are taken from instances contained in `scipy.stats.distributions`.
    So, if the distribution you are searching for is there, then `DimensionBuilder`
    can build one dimension with that prior!

    Examples
    --------
    >>> dimbuilder = DimensionBuilder()
    >>> dimbuilder.build('learning_rate', 'loguniform(0.001, 1, shape=10)')
    Real(name=learning_rate, prior={reciprocal: (0.001, 1), {}}, shape=(10,))
    >>> dimbuilder.build('something_else', 'poisson(mu=3)')
    Integer(name=something_else, prior={poisson: (), {'mu': 3}}, shape=())
    >>> dim = dimbuilder.build('other2', 'uniform(-5, 2)')
    >>> dim
    Real(name=other2, prior={uniform: (-5, 7), {}}, shape=())
    >>> dim.interval()
    (-5.0, 2.0)

    """

    def __init__(self):
        """Init of `DimensionBuilder`."""
        self.name = None

    def choices(self, *args, **kwargs):
        """Create a `Categorical` dimension."""
        name = self.name
        try:
            if isinstance(args[0], (dict, list)):
                return Categorical(name, *args, **kwargs)
        except IndexError as exc:
            raise TypeError("Parameter '{}': "
                            "Expected argument with categories.".format(name)) from exc

        return Categorical(name, args, **kwargs)

    def fidelity(self, *args, **kwargs):
        """Create a `Fidelity` dimension."""
        name = self.name
        return Fidelity(name, *args, **kwargs)

    def uniform(self, *args, **kwargs):
        """Create an `Integer` or `Real` uniformly distributed dimension.

        .. note:: Changes scipy convention for uniform's arguments. In scipy,
           ``uniform(a, b)`` means uniform in the interval [a, a+b). Here, it
           means uniform in the interval [a, b).

        """
        name = self.name
        klass = _real_or_int(kwargs)
        if len(args) == 2:
            return klass(name, 'uniform', args[0], args[1] - args[0], **kwargs)
        return klass(name, 'uniform', *args, **kwargs)

    def gaussian(self, *args, **kwargs):
        """Synonym for `scipy.stats.distributions.norm`."""
        return self.normal(*args, **kwargs)

    def normal(self, *args, **kwargs):
        """Another synonym for `scipy.stats.distributions.norm`."""
        name = self.name
        klass = _real_or_int(kwargs)
        return klass(name, 'norm', *args, **kwargs)

    def loguniform(self, *args, **kwargs):
        """Return a `Dimension` object with
        `scipy.stats.distributions.reciprocal` prior distribution.
        """
        name = self.name
        klass = _real_or_int(kwargs)
        return klass(name, 'reciprocal', *args, **kwargs)

    def _build(self, name, expression):
        """Build a `Dimension` object using a string as its `name` and another
        string, `expression`, from configuration as a "function" to create it.
        """
        self.name = name
        _check_expr_to_eval(expression)

        prior, arg_string = re.findall(r'([a-z][a-z0-9_]*)\((.*)\)', expression)[0]
        globals_ = {'__builtins__': {}}
        try:
            dimension = eval("self." + expression, globals_, {'self': self})

            return dimension
        except AttributeError:
            pass

        # If not found in the methods of `DimensionBuilder`.
        # try to see if it is legit scipy stuff and call a `Dimension`
        # appropriately.
        args, kwargs = eval("_get_arguments(" + arg_string + ")",
                            globals_,
                            {'_get_arguments': _get_arguments})

        if hasattr(sp_dists._continuous_distns, prior):
            klass = _real_or_int(kwargs)
        elif hasattr(sp_dists._discrete_distns, prior):
            klass = Integer
        else:
            raise TypeError("Parameter '{0}': "
                            "'{1}' does not correspond to a supported distribution.".format(
                                name, prior))
        dimension = klass(name, prior, *args, **kwargs)

        return dimension

    def build(self, name, expression):
        """Check `DimensionBuilder._build` for documentation.

        .. note:: Warm-up: Fail early if arguments make object not usable.

        """
        try:
            dimension = self._build(name, expression)
        except ValueError as exc:
            raise TypeError("Parameter '{}': Incorrect arguments.".format(name)) from exc
        except IndexError as exc:
            error_msg = "Parameter '{0}': Please provide a valid form for prior:\n"\
                        "'distribution(*args, **kwargs)'\n"\
                        "Provided: '{1}'".format(name, expression)
            raise TypeError(error_msg) from exc

        try:
            dimension.sample()
        except TypeError as exc:
            error_msg = "Parameter '{0}': Incorrect arguments for distribution '{1}'.\n"\
                        "Scipy Docs::\n\n{2}".format(name,
                                                     dimension._prior_name,
                                                     dimension.prior.__doc__)
            raise TypeError(error_msg) from exc
        except ValueError as exc:
            raise TypeError("Parameter '{0}': Incorrect arguments.".format(name)) from exc

        return dimension


class SpaceBuilder(object):
    """Build a `Space` object form user's configuration."""

    def __init__(self):
        """Initialize a `SpaceBuilder`."""
        self.dimbuilder = DimensionBuilder()
        self.space = None

        self.commands_tmpl = None

        self.converter = None
        self.parser = None

    def build(self, configuration):
        """Create a definition of the problem's search space.

        Using information from the user's script configuration (if provided) and the
        command line arguments, will create a `Space` object defining the problem's
        search space.

        Parameters
        ----------
        configuration: OrderedDict
            An OrderedDict containing the name and the expression of the parameters.

        Returns
        -------
        `orion.algo.space.Space`
            The problem's search space definition.

        """
        self.space = Space()
        for namespace, expression in flatten(configuration).items():
            if _should_not_be_built(expression):
                continue

            expression = _remove_marker(expression)
            dimension = self.dimbuilder.build(namespace, expression)

            try:
                self.space.register(dimension)
            except ValueError as exc:
                error_msg = 'Conflict for name \'{}\' in parameters'.format(namespace)
                raise ValueError(error_msg) from exc

        return self.space

    def build_to(self, config_path, trial, experiment=None):
        """Create the configuration for the user's script.

        Using the configuration parser, create the commandline associated with the
        user's script while replacing the correct instances of parameter distributions by
        their actual values. If needed, the parser will also create a configuration file.

        Parameters
        ----------
        config_path: str
            Path in which the configuration file instance will be created.
        trial: `orion.core.worker.trial.Trial`
            Object with concrete parameter values for the defined `Space`.
        experiment: `orion.core.worker.experiment.Experiment`, optional
            Object with information related to the current experiment.

        Returns
        -------
        list
            The commandline arguments that must be given to script for execution.

        """
        return self.parser.format(config_path, trial, experiment)
